import os
import unittest

import numpy as np
from numpy.testing import assert_array_almost_equal

import skrf as rf


class CircuitTestConstructor(unittest.TestCase):
    """
    Various tests on the Circuit constructor.
    """
    def setUp(self):
        # Importing network examples
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'
        self.ntwk1 = rf.Network(os.path.join(self.test_dir, 'ntwk1.s2p'))
        self.ntwk2 = rf.Network(os.path.join(self.test_dir, 'ntwk2.s2p'))
        self.freq = self.ntwk1.frequency
        # circuit external ports
        self.port1 = rf.Circuit.Port(self.freq, name='Port1')
        self.port2 = rf.Circuit.Port(self.freq, name='Port2')
        # circuit connections
        self.connections = [[(self.port1, 0), (self.ntwk1, 0)],
                       [(self.ntwk1, 1), (self.ntwk2, 0)],
                       [(self.ntwk2, 1), (self.port2, 0)]]
        self.circuit = rf.Circuit(self.connections)

    def test_all_networks_have_name(self):
        """
        Check that a Network without name raises an exception
        """
        _ntwk1 = self.ntwk1.copy()
        connections = [[(self.port1, 0), (_ntwk1, 0)],
                       [(_ntwk1, 1), (self.ntwk2, 0)],
                       [(self.ntwk2, 1), (self.port2, 0)]]

        _ntwk1.name = []
        self.assertRaises(AttributeError, rf.Circuit, connections)

        _ntwk1.name = ''
        self.assertRaises(AttributeError, rf.Circuit, connections)

    def test_all_networks_have_same_frequency(self):
        """
        Check that a Network with a different frequency than the other
        raises an exception
        """
        _ntwk1 = self.ntwk1.copy()
        connections = [[(self.port1, 0), (_ntwk1, 0)],
                       [(_ntwk1, 1), (self.ntwk2, 0)],
                       [(self.ntwk2, 1), (self.port2, 0)]]

        _ntwk1.frequency = rf.Frequency(start=1, stop=1, npoints=1, unit='GHz')
        self.assertRaises(AttributeError, rf.Circuit, connections)

    def test_no_duplicate_node(self):
        """
        Check that a circuit description has no duplicated (network, port)
        """
        # (port1, 0) is found twice in the connections description
        connections = [[(self.port1, 0), (self.ntwk1, 0)],
                       [(self.ntwk1, 1), (self.ntwk2, 0)],
                       [(self.ntwk2, 1), (self.port1, 0)]]
        self.assertRaises(AttributeError, rf.Circuit, connections)

    def test_s_active(self):
        """
        Test the active s-parameter of a 2-ports network
        """
        # s_act should be equal to s11 if a = [1,0]
        assert_array_almost_equal(self.circuit.s_active([1, 0])[:,0], self.circuit.s_external[:,0,0])
        # s_act should be equal to s22 if a = [0,1]
        assert_array_almost_equal(self.circuit.s_active([0, 1])[:,1], self.circuit.s_external[:,1,1])

    def test_auto_reduce(self):
        """
        Test the auto_reduce parameter of the Circuit constructor
        """
        # Explicitly enable the `reduce_circuit` method
        reduced_circuit = rf.Circuit(self.connections, auto_reduce=True)
        assert_array_almost_equal(self.circuit.s_external, reduced_circuit.s_external)
        self.assertNotEqual(self.circuit.dim, reduced_circuit.dim)

        # Implicitly enable the `reduce_circuit` method
        implicit_reduced_circuit = rf.Circuit(self.connections, max_nports=10)
        self.assertEqual(reduced_circuit.dim, implicit_reduced_circuit.dim)

    def test_auto_reduce_with_passed_arguments(self):
        """
        Test the auto_reduce parameter of the Circuit constructor with passed arguments
        """
        # Test with max_nports=1 and dynamic_networks in a tuple or a list
        kwargs = [('max_nports', 1),
                  ('dynamic_networks', (self.ntwk2,)),
                  ('dynamic_networks', [self.ntwk2])]

        # No connections should be reduced
        for key, value in kwargs:
            circuit = rf.Circuit(self.connections, **{key: value})
            assert_array_almost_equal(self.circuit.s_external, circuit.s_external)
            self.assertEqual(circuit.connections, self.connections)

    def test_update_networks(self):
        """
        Test the update_networks method of the Circuit class
        """
        # Create an abnormal circuit with a random network
        ntwk3 = self.ntwk1.copy()
        ntwk3.name = 'ntwk3'
        connections = [[(self.port1, 0), (self.ntwk1, 0)],
                       [(self.ntwk1, 1), (ntwk3, 0)],
                       [(ntwk3, 1), (self.port2, 0)]]
        circuit = rf.Circuit(connections, dynamic_networks=(ntwk3,))

        # should raise an exception if passing a network not in the circuit
        with self.assertRaises(ValueError):
            _ = circuit.update_networks(networks=(self.ntwk2,))

        # Update the circuit with the self.ntwk2
        ntwk3.s = self.ntwk2.s

        # Check the result type and values
        circuit_updated = circuit.update_networks(networks=(ntwk3,))
        self.assertTrue(isinstance(circuit_updated, rf.Circuit))
        assert_array_almost_equal(self.circuit.s_external, circuit_updated.s_external)

        # Check the result type and values when updating the circuit inplace
        none_result = circuit.update_networks(networks=(ntwk3,), inplace=True)
        self.assertIsNone(none_result)
        assert_array_almost_equal(self.circuit.s_external, circuit.s_external)

    def test_cache_attributes(self):
        """
        Test the cached attributes of the Circuit
        """
        init_circuit = rf.Circuit(self.connections)

        cached_attributes = rf.Circuit.CACHEDPROPERTIES

        # Initial circuit should not have cached attributes
        for attr in cached_attributes:
            self.assertFalse(init_circuit.__dict__.get(attr, False))

        # Access the attributes to cache them
        values = [getattr(init_circuit, attr) for attr in cached_attributes]

        # Check that the cached attributes are set correctly
        for attr, value in zip(cached_attributes, values):
            assert_array_almost_equal(init_circuit.__dict__.get(attr, 0.0), value)

        # Modify connections to invalidate the cache
        init_circuit.connections = self.connections
        for attr in cached_attributes:
            self.assertFalse(init_circuit.__dict__.get(attr, False))

    def test_auxiliary_matrix(self):
        """
        Test the auxiliary matrix [T] of the Circuit
        """
        # Calculate the auxiliary matrix T and C @ X
        X, C = self.circuit.X, self.circuit.C
        X_F, C_F = self.circuit.X_F, self.circuit.C_F
        rst, T = - C @ X, self.circuit.T

        # Check that auxiliary_matrices is equal to initial values
        assert_array_almost_equal(T, rst)
        assert_array_almost_equal(X, X_F)
        assert_array_almost_equal(C, C_F)

        # Check the memory layout of the matrices
        self.assertTrue(rst.flags['C_CONTIGUOUS'])
        self.assertTrue(T.flags['F_CONTIGUOUS'])
        self.assertTrue(self.circuit.C.flags['C_CONTIGUOUS'])
        self.assertTrue(self.circuit.C_F.flags['F_CONTIGUOUS'])
        self.assertTrue(self.circuit.X.flags['C_CONTIGUOUS'])
        self.assertTrue(self.circuit.X_F.flags['F_CONTIGUOUS'])

class CircuitClassMethods(unittest.TestCase):
    """
    Test the various class methods of Circuit such as Ground, Port, etc.
    """
    def setUp(self):
        self.freq = rf.Frequency(start=1, stop=2, npoints=101, unit='GHz')
        self.media = rf.DefinedGammaZ0(self.freq)

    def test_ground(self):
        """
        Ground object are infinite shunt admittance (ie. a 2-port short)
        """
        # should raise an exception if no name is passed
        with self.assertRaises(TypeError):
            gnd = rf.Circuit.Ground(self.freq)

        gnd = rf.Circuit.Ground(self.freq, 'gnd')
        gnd_ref = rf.Network(frequency=self.freq,
                             s=np.tile(np.array([[-1,]]),
                                       (len(self.freq),1,1)))

        assert_array_almost_equal(gnd.s, gnd_ref.s)


    def test_open(self):
        """
        Open object are infinite series resistance (ie. a 2-port open)
        """
        # should raise an exception if no name is passed
        with self.assertRaises(TypeError):
            opn = rf.Circuit.Open(self.freq)

        opn = rf.Circuit.Open(self.freq, 'open')
        opn_ref = rf.Network(frequency=self.freq,
                             s=np.tile(np.array([[1]]),
                                       (len(self.freq),1,1)))

        assert_array_almost_equal(opn.s, opn_ref.s)

    def test_series_impedance(self):
        z0s = [1, 50]
        Zs = [1, 1 + 1j, rf.INF]
        for z0 in z0s:
            for Z in Zs:
                assert_array_almost_equal(
                    rf.Circuit.SeriesImpedance(self.freq, Z, 'imp', z0=z0).s,
                    self.media.resistor(Z, z0=z0).s
                    )

        # Z=0 is a thru
        assert_array_almost_equal(
            rf.Circuit.SeriesImpedance(self.freq, Z=0, name='imp').s,
            self.media.thru().s
            )

    def test_shunt_admittance(self):
        z0s = [1, 50]
        Ys = [1, 1 + 1j, rf.INF]
        for z0 in z0s:
            for Y in Ys:
                assert_array_almost_equal(
                    rf.Circuit.ShuntAdmittance(self.freq, Y, 'imp', z0=z0).s,
                    self.media.shunt(self.media.load(rf.zl_2_Gamma0(z0, 1/Y))).s
                    )

        # Y=INF is a a 2-ports short
        assert_array_almost_equal(
            rf.Circuit.ShuntAdmittance(self.freq, rf.INF, 'imp').s,
            self.media.short(nports=2).s
            )

class CircuitTestWilkinson(unittest.TestCase):
    """
    Create a Wilkinson power divider Circuit [#]_ and test the results
    against network.connect.
    The results in [#]_ do not agree due to an error in the formula (3)
    for mismatched intersections.

    References
    ----------
    .. [#] https://en.wikipedia.org/wiki/Wilkinson_power_divider
    .. [#] P. Hallbjörner, Microw. Opt. Technol. Lett. 38, 99 (2003).


    """
    def setUp(self):
        """
        Circuit setup
        """
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'
        self.freq = rf.Frequency(start=1, stop=2, npoints=101, unit='GHz')
        # characteristic impedance of the ports
        Z0_ports = 50

        # resistor
        self.R = 100
        self.line_resistor = rf.media.DefinedGammaZ0(frequency=self.freq, z0=Z0_ports)
        self.resistor = self.line_resistor.resistor(self.R, name='resistor')

        # branches
        Z0_branches = np.sqrt(2)*Z0_ports
        self.line_branches = rf.media.DefinedGammaZ0(frequency=self.freq, z0=Z0_branches)
        self.branch1 = self.line_branches.line(90, unit='deg', name='branch1')
        self.branch2 = self.line_branches.line(90, unit='deg', name='branch2')

        # ports
        port1 = rf.Circuit.Port(self.freq, name='port1', z0=50)
        port2 = rf.Circuit.Port(self.freq, name='port2', z0=50)
        port3 = rf.Circuit.Port(self.freq, name='port3', z0=50)

        # Connection setup
        self.connections = [
                   [(port1, 0), (self.branch1, 0), (self.branch2, 0)],
                   [(port2, 0), (self.branch1, 1), (self.resistor, 0)],
                   [(port3, 0), (self.branch2, 1), (self.resistor, 1)]
                ]

        self.C = rf.Circuit(self.connections)

        # theoretical results from ref P.Hallbjörner (2003)
        self.X1_nn = np.array([1 - np.sqrt(2), -1, -1])/(1 + np.sqrt(2))
        self.X2_nn = np.array([1 - np.sqrt(2), -3 + np.sqrt(2),
                               -1 - np.sqrt(2)])/(3 + np.sqrt(2))

        self.X1_m1 = 2 / (1 + np.sqrt(2))
        self.X1_m2 = np.sqrt(2) / (1 + np.sqrt(2))
        self.X1_m3 = np.sqrt(2) / (1 + np.sqrt(2))

        self.X2_m1 = 4 / (3 + np.sqrt(2))
        self.X2_m2 = 2*np.sqrt(2) / (3 + np.sqrt(2))
        self.X2_m3 = 2 / (3 + np.sqrt(2))

        self.X1 = np.array([[0, self.X1_m2, self.X1_m3],
                            [self.X1_m1, 0, self.X1_m3],
                            [self.X1_m1, self.X1_m2, 0]]) + np.diag(self.X1_nn)
        self.X2 = np.array([[0, self.X2_m2, self.X2_m3],
                            [self.X2_m1, 0, self.X2_m3],
                            [self.X2_m1, self.X2_m2, 0]]) + np.diag(self.X2_nn)


    @unittest.expectedFailure
    def test_sparam_individual_intersection_matrices(self):
        """
        Testing the individual intersection scattering matrices X_k
        The results in [#]_ do not agree due to an error in the formula (3)
        for mismatched intersections.

        References
        ----------
        .. [#] P. Hallbjörner, Microw. Opt. Technol. Lett. 38, 99 (2003).
        """
        np.testing.assert_array_almost_equal(self.C._Xk(self.connections[0])[0], self.X1)
        np.testing.assert_array_almost_equal(self.C._Xk(self.connections[1])[0], self.X2)
        np.testing.assert_array_almost_equal(self.C._Xk(self.connections[2])[0], self.X2)

    @unittest.expectedFailure
    def test_sparam_global_intersection_matrix(self):
        """
        Testing the global intersection scattering matrix
        The results in [#]_ do not agree due to an error in the formula (3)
        for mismatched intersections.

        References
        ----------
        .. [#] P. Hallbjörner, Microw. Opt. Technol. Lett. 38, 99 (2003).
        """
        from scipy.linalg import block_diag
        assert_array_almost_equal(self.C.X[0], block_diag(self.X1, self.X2, self.X2) )

    def test_sparam_circuit(self):
        """
        Testing the external scattering matrix
        """
        S_theoretical = np.array([[0, 1, 1],
                                  [1, 0, 0],
                                  [1, 0, 0]]) * (-1j/np.sqrt(2))

        # extracting the external ports
        S_ext = self.C.s_external

        assert_array_almost_equal(S_ext[0], S_theoretical)

    def test_compare_with_skrf_wilkison(self):
        """
        Create a Wilkinson power divider using skrf usual Network methods.
        """
        z0_port = 50
        # require to create the three tees
        T0 = self.line_branches.splitter(3, z0=z0_port)
        T1 = self.line_branches.splitter(3, z0=z0_port)
        T2 = self.line_branches.splitter(3, z0=z0_port)

        _wilkinson1 = rf.connect(T0, 1, self.branch1, 0)
        _wilkinson2 = rf.connect(_wilkinson1, 2, self.branch2, 0)
        _wilkinson3 = rf.connect(_wilkinson2, 1, T1, 0)
        _wilkinson4 = rf.connect(_wilkinson3, 1, T2, 0)
        _wilkinson5 = rf.connect(_wilkinson4, 1, self.resistor, 0)
        wilkinson = rf.innerconnect(_wilkinson5, 1, 4)

        ntw_C = self.C.network

        assert_array_almost_equal(ntw_C.s, wilkinson.s)
        assert_array_almost_equal(ntw_C.z0, wilkinson.z0)

    def test_compare_with_designer_wilkinson(self):
        """
        Compare the result with ANSYS Designer model

        Built as in https://www.microwaves101.com/encyclopedias/wilkinson-power-splitters
        """
        designer_wilkinson = rf.Network(os.path.join(self.test_dir, 'designer_wilkinson_splitter.s3p'))
        ntw_C = self.C.network

        assert_array_almost_equal(ntw_C.s[0], designer_wilkinson.s[0], decimal=4)

    def test_s_active(self):
        """
        Test the active s-parameter of a 3-ports network
        """
        # s_act should be equal to s11 if a = [1,0,0]
        assert_array_almost_equal(self.C.network.s_active([1, 0, 0])[:,0], self.C.s_external[:,0,0])
        # s_act should be equal to s22 if a = [0,1,0]
        assert_array_almost_equal(self.C.network.s_active([0, 1, 0])[:,1], self.C.s_external[:,1,1])
        # s_act should be equal to s33 if a = [0,0,1]
        assert_array_almost_equal(self.C.network.s_active([0, 0, 1])[:,2], self.C.s_external[:,2,2])

    def test_circuit_reduce_with_split_multi(self):
        """
        Test the reduce_circuit method with split_multi=True.
        """
        ntw_C = self.C.network

        C_reduced = rf.Circuit(self.connections, split_multi=True)
        ntw_C_reduced = C_reduced.network

        self.assertNotEqual(self.C.dim, C_reduced.dim)
        assert_array_almost_equal(ntw_C.s, ntw_C_reduced.s)

class CircuitTestCascadeNetworks(unittest.TestCase):
    """
    Build a circuit made of two Networks cascaded and compare the result
    to usual cascading of two networks.
    """
    def setUp(self):
        # Importing network examples
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'
        self.ntwk1 = rf.Network(os.path.join(self.test_dir, 'ntwk1.s2p'))
        self.ntwk2 = rf.Network(os.path.join(self.test_dir, 'ntwk2.s2p'))
        # ntwk3 is the cascade of ntwk1 and ntwk2, ie. self.ntwk1 ** self.ntwk2
        self.ntwk3 = rf.Network(os.path.join(self.test_dir, 'ntwk3.s2p'))
        self.freq = self.ntwk1.frequency
        # circuit external ports
        self.port1 = rf.Circuit.Port(self.freq, name='Port1')
        self.port2 = rf.Circuit.Port(self.freq, name='Port2')
        self.open = rf.Circuit.Open(self.freq, name='PortO')
        _media = rf.media.DefinedGammaZ0(frequency=self.freq)
        self.match = _media.match(name='match')

    def test_cascade(self):
        """
        Compare ntwk3 to the Circuit of ntwk1 and ntwk2.
        """
        connections = [  [(self.port1, 0), (self.ntwk1, 0)],
                         [(self.ntwk1, 1), (self.ntwk2, 0)],
                         [(self.ntwk2, 1), (self.port2, 0)] ]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(circuit.s_external, self.ntwk3.s)

    def test_cascade2(self):
        """
        Same thing with different ordering of the connections.
        Demonstrate that changing the connections setup order does not change
        the result.
        """
        connections = [  [(self.port1, 0), (self.ntwk1, 0)],
                         [(self.ntwk2, 0), (self.ntwk1, 1)],
                         [(self.port2, 0), (self.ntwk2, 1)] ]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(circuit.s_external, self.ntwk3.s)

    def test_cascade3(self):
        """
        Inverting the cascading network order
        Demonstrate that changing the connections setup order does not change
        the result (at the requirement that port impedance are the same).
        """
        connections = [  [(self.port1, 0), (self.ntwk2, 0)],
                         [(self.ntwk2, 1), (self.ntwk1, 0)],
                         [(self.port2, 0), (self.ntwk1, 1)] ]
        circuit = rf.Circuit(connections)
        ntw = self.ntwk2 ** self.ntwk1
        assert_array_almost_equal(circuit.s_external, ntw.s)

    def test_open_condensation(self):
        """
        Test the impact of omitting an open termination in the circuit connections.
        Compare the result of a conventional circuit setup with an open termination
        to a condensed setup that assumes an open termination by default.
        """
        # Conventional setup including open termination
        cnx_con = [  [(self.port1, 0), (self.ntwk1, 0)],
                     [(self.ntwk1, 1), (self.ntwk2, 0), (self.port2, 0)],
                     [(self.ntwk2, 1), (self.open, 0)] ]
        ckt_con = rf.Circuit(cnx_con)

        # Condensed setup where open termination is implied
        cnx_cds = [  [(self.port1, 0), (self.ntwk1, 0)],
                     [(self.ntwk1, 1), (self.ntwk2, 0), (self.port2, 0)],
                     [(self.ntwk2, 1)] # Open termination is implied
                     ]

        ckt_cds = rf.Circuit(cnx_cds)

        assert_array_almost_equal(ckt_con.s_external, ckt_cds.s_external)

    def test_match_condensation(self):
        """
        Test the impact of omitting a matching network in the circuit connections.
        Compare the result of a conventional circuit setup with a matching network
        to a condensed setup that assumes the matching network by default.
        """
        # Conventional setup including matching network
        cnx_con = [  [(self.port1, 0), (self.ntwk1, 0)],
                     [(self.ntwk1, 1), (self.ntwk2, 0), (self.port2, 0)],
                     [(self.ntwk2, 1), (self.match, 0)] ]
        ckt_con = rf.Circuit(cnx_con)

        # Condensed setup where matching network is implied
        cnx_cds = [  [(self.port1, 0), (self.ntwk1, 0)],
                     [(self.ntwk1, 1), (self.ntwk2, 0), (self.port2, 0)]
                     ]

        ckt_cds = rf.Circuit(cnx_cds)

        assert_array_almost_equal(ckt_con.s_external, ckt_cds.s_external)

class CircuitTestMultiPortCascadeNetworks(unittest.TestCase):
    """
    Various 1-ports, 2-ports and 4-ports circuits and associated tests
    """
    def test_1port_matched_load(self):
        """
        Connect a matched load directly to the port
        """
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        port1 = rf.Circuit.Port(freq,  name='port1')
        line = rf.media.DefinedGammaZ0(frequency=freq)
        match_load = line.match(name='match_load')

        cnx = [
            [(port1, 0), (match_load, 0)]
        ]
        cir = rf.Circuit(cnx)

        assert_array_almost_equal(match_load.s, cir.s_external)

    def test_1port_short(self):
        """
        Connect a short directly to the port
        """
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        port1 = rf.Circuit.Port(freq,  name='port1')
        line = rf.media.DefinedGammaZ0(frequency=freq)
        short = line.short(name='short')
        gnd1 = rf.Circuit.Ground(freq, name='gnd')
        # method 1 : use the Ground Network (which 2 port actually)
        cnx = [
            [(port1, 0), (gnd1, 0)]
        ]
        cir = rf.Circuit(cnx)
        assert_array_almost_equal(short.s, cir.s_external)
        # method 2 : use a short Network (1 port)
        cnx = [
            [(port1, 0), (short, 0)]
        ]
        cir = rf.Circuit(cnx)

        assert_array_almost_equal(short.s, cir.s_external)

    def test_1port_random_load(self):
        """
        Connect a random load directly to the port
        """
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        port1 = rf.Circuit.Port(freq,  name='port1')
        line = rf.media.DefinedGammaZ0(frequency=freq)
        gamma = np.random.default_rng().random((1,1)) + 1j*np.random.default_rng().random((1,1))
        load = line.load(gamma, name='load')

        cnx = [
            [(port1, 0), (load, 0)]
        ]
        cir = rf.Circuit(cnx)

        assert_array_almost_equal(load.s, cir.s_external)

    def test_1port_matched_network_default_impedance(self):
        """
        Connect a random 2 port network connected to a matched load
        """
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.default_rng().random(4).reshape(2, 2)
        line = rf.media.DefinedGammaZ0(frequency=freq)
        match_load = line.match(name='match_load')

        # classic connecting
        b = a ** match_load

        # Circuit connecting
        port1 = rf.Circuit.Port(freq,  name='port1')
        connections = [[(port1, 0), (a, 0)], [(a, 1), (match_load, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(b.s, circuit.s_external)

    def test_1port_matched_network_complex_impedance(self):
        """
        Connect a 2 port network to a complex impedance.
        Both ports are complex.
        """
        z01, z02 = 1-1j, 2+4j
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.default_rng().random(4).reshape(2, 2)
        a.z0 = [z01, z02]
        line = rf.media.DefinedGammaZ0(frequency=freq, z0=z02)
        match_load = line.match(name='match_load')

        # classic connecting
        b = a ** match_load

        # Circuit connecting
        port1 = rf.Circuit.Port(freq, z0=z01, name='port1')
        connections = [[(port1, 0), (a,0)], [(a, 1), (match_load, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(b.s, circuit.s_external)

    def test_2ports_default_characteristic_impedance(self):
        """
        Connect two 2-ports networks in a resulting  2-ports network,
        same default charact impedance (50 Ohm) for all ports
        """
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.default_rng().random(4).reshape(2, 2)

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.default_rng().random(4).reshape(2, 2)

        # classic connecting
        c = rf.connect(a, 1, b, 0)

        # Circuit connecting
        port1 = rf.Circuit.Port(freq,  name='port1')
        port2 = rf.Circuit.Port(freq,  name='port2')

        connections = [[(port1, 0), (a, 0)],
                       [(a, 1), (b, 0)],
                       [(b, 1), (port2, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_2ports_complex_characteristic_impedance(self):
        """
        Connect two 2-ports networks in a resulting  2-ports network,
        same complex charact impedance (1+1j) for all ports
        """
        z0 = 1 + 1j
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.default_rng().random(4).reshape(2, 2)
        a.z0 = z0

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.default_rng().random(4).reshape(2, 2)
        b.z0 = z0

        # classic connecting
        c = rf.connect(a, 1, b, 0)

        # Circuit connecting
        port1 = rf.Circuit.Port(freq, z0=z0, name='port1')
        port2 = rf.Circuit.Port(freq, z0=z0, name='port2')

        connections = [[(port1, 0), (a, 0)],
                       [(a, 1), (b, 0)],
                       [(b, 1), (port2, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_2ports_different_characteristic_impedances(self):
        """
        Connect two 2-ports networks in a resulting  2-ports network,
        different characteristic impedances for each network ports
        """
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.default_rng().random(4).reshape(2,2)
        a.z0 = [1, 2]  #  Z0 should never be zero

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.default_rng().random(4).reshape(2,2)
        b.z0 = [11, 12]

        # classic connecting
        c = rf.connect(a, 1, b, 0)

        # Circuit connecting
        port1 = rf.Circuit.Port(freq, z0=1, name='port1')
        port2 = rf.Circuit.Port(freq, z0=12, name='port2')

        connections = [[(port1, 0), (a, 0)],
                       [(a, 1), (b, 0)],
                       [(b, 1), (port2, 0)] ]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_4ports_default_characteristic_impedances(self):
        """
        Connect two 4-ports networks in a resulting 4-ports network,
        with default characteristic impedances
        """
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.default_rng().random(16).reshape(4, 4)

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.default_rng().random(16).reshape(4, 4)

        # classic connecting
        c = rf.connect(a, 2, b, 0, 2)

        # Circuit connecting
        port1 = rf.Circuit.Port(freq, name='port1')
        port2 = rf.Circuit.Port(freq, name='port2')
        port3 = rf.Circuit.Port(freq, name='port3')
        port4 = rf.Circuit.Port(freq, name='port4')

        connections = [[(port1, 0), (a, 0)],
                       [(port2, 0), (a, 1)],
                       [(a, 2), (b, 0)],
                       [(a, 3), (b, 1)],
                       [(b, 2), (port3, 0)],
                       [(b, 3), (port4, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_4ports_complex_characteristic_impedances(self):
        """
        Connect two 4-ports networks in a resulting 4-ports network,
        with same complex characteristic impedances
        """
        z0 = 5 + 4j
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.default_rng().random(16).reshape(4, 4)
        a.z0 = z0

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.default_rng().random(16).reshape(4, 4)
        b.z0 = z0

        # classic connecting
        c = rf.connect(a, 2, b, 0, 2)

        # circuit connecting
        port1 = rf.Circuit.Port(freq, z0=z0, name='port1')
        port2 = rf.Circuit.Port(freq, z0=z0, name='port2')
        port3 = rf.Circuit.Port(freq, z0=z0, name='port3')
        port4 = rf.Circuit.Port(freq, z0=z0, name='port4')

        connections = [ [(port1, 0), (a, 0)],
                        [(port2, 0), (a, 1)],
                        [(a, 2), (b, 0)],
                        [(a, 3), (b, 1)],
                        [(b, 2), (port3, 0)],
                        [(b, 3), (port4, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_4ports_different_characteristic_impedances(self):
        """
        Connect two 4-ports networks in a resulting 4-ports network,
        with different characteristic impedances
        """
        z0 = [1, 2, 3, 4]
        freq = rf.Frequency(start=1, npoints=1, unit='GHz')
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.default_rng().random(16).reshape(4, 4)
        a.z0 = z0

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.default_rng().random(16).reshape(4, 4)
        b.z0 = [11, 12, 13, 14]

        # classic connecting
        _c = rf.connect(a, 2, b, 0)
        c = rf.innerconnect(_c, 2, 3)

        # Circuit connecting
        port1 = rf.Circuit.Port(freq, z0=1, name='port1')
        port2 = rf.Circuit.Port(freq, z0=2, name='port2')
        port3 = rf.Circuit.Port(freq, z0=13, name='port3')
        port4 = rf.Circuit.Port(freq, z0=14, name='port4')

        connections = [[(port1, 0), (a, 0)],
                       [(port2, 0), (a, 1)],
                       [(a, 2), (b, 0)],
                       [(a, 3), (b, 1)],
                       [(b, 2), (port3, 0)],
                       [(b, 3), (port4, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_shunt_element(self):
        """
        Compare a shunt element network (here a capacitor)
        """
        freq = rf.Frequency(start=1, stop=2, npoints=101, unit='GHz')
        line = rf.media.DefinedGammaZ0(frequency=freq, z0=50)
        # usual way
        cap_shunt_manual = line.shunt_capacitor(50e-12)

        # A Circuit way
        port1 = rf.Circuit.Port(frequency=freq, name='port1', z0=50)
        port2 = rf.Circuit.Port(frequency=freq, name='port2', z0=50)
        cap_shunt = line.capacitor(50e-12, name='cap_shunt')
        ground = rf.Circuit.Ground(frequency=freq, name='ground', z0=50)

        connections = [
            [(port1, 0), (cap_shunt, 0), (port2, 0)],
            [(cap_shunt, 1), (ground, 0)]
        ]

        # # Another possibility could have been without ground :
        # shunt_cap = line.shunt_capacitor(50e-12)
        # shunt_cap.name='shunt_cap'
        # connections = [
        #     [(port1, 0), (shunt_cap, 0)],
        #     [(shunt_cap ,1), (port2, 0)]
        # ]

        cap_shunt_from_circuit = rf.Circuit(connections).network

        assert_array_almost_equal(cap_shunt_manual.s, cap_shunt_from_circuit.s)

    def test_reduce_circuit(self):
        """
        Test reducing multi-port cascade circuit
        """
        freq_n = 101
        freq = rf.Frequency(start=1, stop=2, npoints=freq_n, unit='GHz')
        line = rf.media.DefinedGammaZ0(frequency=freq, z0=50)

        # network A
        ntwkA = rf.Network(name='a')
        ntwkA.frequency = freq
        ntwkA_z0 = (60, 70, 80, 90, 100)
        ntwkA_np = len(ntwkA_z0)
        ntwkA.z0 = [ntwkA_z0]*freq_n
        ntwkA.s = np.random.default_rng().random(freq_n * ntwkA_np**2).reshape(freq_n, ntwkA_np, ntwkA_np)

        # network B
        ntwkB = rf.Network(name='b')
        ntwkB.frequency = freq
        ntwkB_z0 = (10, 20, 30, 40, 50)
        ntwkB_np = len(ntwkB_z0)
        ntwkB.z0 = [ntwkB_z0]*freq_n
        ntwkB.s = np.random.default_rng().random(freq_n * ntwkB_np**2).reshape(freq_n, ntwkB_np, ntwkB_np)

        # Construct the connection
        port1 = rf.Circuit.Port(frequency=freq, name='port1', z0=50)
        port2 = rf.Circuit.Port(frequency=freq, name='port2', z0=50)
        ground1 = rf.Circuit.Ground(frequency=freq, name='ground1', z0=50)
        ground2 = rf.Circuit.Ground(frequency=freq, name='ground2', z0=50)
        open_port = rf.Circuit.Open(frequency=freq, name='open_port', z0=50)
        ind_shunt = line.inductor(50e-12, name='ind_shunt')
        cap_series = line.capacitor(30e-9, name='cap_series')
        port3 = rf.Circuit.Port(frequency=freq, name='port3', z0=50)
        port4 = rf.Circuit.Port(frequency=freq, name='port4', z0=50)

        connections = [
            [(ntwkA, 0), (cap_series, 0)],
            [(port1, 0), (ind_shunt, 0), (cap_series, 1)],
            [(ground1, 0), (ind_shunt, 1)],
            [(ntwkA, 1), (port2, 0)],
            [(ntwkA, 2), (ground2, 0)],
            [(ntwkA, 3), (open_port, 0)],
            [(ntwkA, 4), (ntwkB, 0)],
            [(ntwkB, 1), (port3, 0)],
            [(ntwkB, 2), (port4, 0)],
            [(ntwkB, 3), (ntwkB, 4)],
        ]

        # Circuit connecting
        circuit = rf.Circuit(connections)

        # Reduce the circuit
        reduced_cnxs = rf.reduce_circuit(connections)
        reduced_circuit = rf.Circuit(reduced_cnxs)

        assert_array_almost_equal(circuit.network.s, reduced_circuit.network.s)

class CircuitTestVariableCoupler(unittest.TestCase):
    """
    If we use 3 dB hybrid defined as :
                   ________
    Input     0 --|       |-- 1 Through
                  |       |
    Isolated  3 --|_______|-- 2 Coupled

    And if we connect two 3 dB hybrid with a phase shifter such as:
                 _________                                 _________
    Port#3 >- 0-|        |-1 -------------------------- 0-|        |-1 -< Port#2
                |  Hyb1  |                                |  Hyb2  |
    Port#0 >- 3-|________|-2 -- 0-[phase shifter]-1 -- 3-|________|-2 -< Port#1

    Then we have a variable coupler. The coupling factor can be adjusted
    by changing the phase of the phase shifter.

    The port order in this example is voluntary complicated to make a good
    example.

    """
    def setUp(self):
        self.freq = rf.Frequency(start=1.5, stop=1.5, npoints=1, unit='GHz')
        self.coax = rf.media.DefinedGammaZ0(frequency=self.freq)
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'

    def phase_shifter(self, phase_deg):
        return self.coax.line(d=phase_deg, unit='deg')

    def hybrid(self, name='hybrid'):
        Sc = 1/np.sqrt(2)*np.array([[ 0,  1, 1j,  0],
                                    [ 1,  0,  0, 1j],
                                    [1j,  0,  0,  1],
                                    [ 0, 1j,  1,  0]])
        hybrid = rf.Network(frequency=self.freq, s=Sc, name=name)
        return hybrid

    def variable_coupler_network_from_connect(self, phase_deg):
        ps = self.phase_shifter(phase_deg)
        hybrid1, hybrid2 = self.hybrid(), self.hybrid()
        # connecting the networks together.
        # NB: in order to know which port corresponds to which, it is usefull
        # to define different port characteristic impedances like
        # hybrid1.z0 = [11, 12, 13, 14]
        # hybrid2.z0 = [21, 22, 23, 24]
        # ps.z0 = [31, 32]
        # and to make a drawing of each steps.
        # This is not convenient, that's why the Circuit approach can be easier.
        _temp = rf.connect(hybrid1, 2, ps, 0)
        _temp = rf.connect(_temp, 1, hybrid2, 0)
        _temp = rf.innerconnect(_temp, 1, 5)
        # re-order port numbers to match the example
        _temp.renumber([0, 1, 2, 3], [3, 0, 2, 1])
        return _temp

    def variable_coupler_circuit(self, phase_deg):
        ps = self.phase_shifter(phase_deg)
        ps.name = 'ps'  # do not forget the name of the network !
        hybrid1, hybrid2 = self.hybrid('hybrid1'), self.hybrid('hybrid2')

        port1 = rf.Circuit.Port(ps.frequency, 'port1')
        port2 = rf.Circuit.Port(ps.frequency, 'port2')
        port3 = rf.Circuit.Port(ps.frequency, 'port3')
        port4 = rf.Circuit.Port(ps.frequency, 'port4')
        # Note that the order of port appearance is important.
        # 1st port to appear in the connection setup will be the 1st port (0),
        # then second to appear the second port (1), etc...
        # There is no constraint for the order of the connections.
        connections = [
                       [(port1, 0), (hybrid1, 3)],
                       [(port2, 0), (hybrid2, 2)],
                       [(port3, 0), (hybrid2, 1)],
                       [(port4, 0), (hybrid1, 0)],
                       [(hybrid1, 2), (ps, 0)],
                       [(hybrid1, 1), (hybrid2, 0)],
                       [(ps, 1), (hybrid2, 3)],
                      ]

        return rf.Circuit(connections)

    def variable_coupler_network_from_circuit(self, phase_deg):
        return self.variable_coupler_circuit(phase_deg).network

    def test_compare_with_network_connect(self):
        """
        Compare with the S-parameters obtained from Network.connect
        """
        phase_deg = np.random.default_rng().integers(low=0, high=180)
        vc_connect = self.variable_coupler_network_from_connect(phase_deg)
        vc_circuit = self.variable_coupler_network_from_circuit(phase_deg)
        assert_array_almost_equal(vc_connect.s, vc_circuit.s)

    def test_compare_with_designer(self):
        """
        Compare with the S-parameters obtained from ANSYS Designer
        """
        for phase_angle in [20, 75]:
            vc_designer = rf.Network(os.path.join(self.test_dir,
                                                  f"designer_variable_coupler_ideal_{phase_angle}deg.s4p"))
            vc_circuit = self.variable_coupler_network_from_circuit(phase_angle)
            assert_array_almost_equal(vc_designer.s, vc_circuit.s, decimal=4)

    def test_compare_connect_and_designer(self):
        """
        Compare S-parameters obtained from ANSYS Designer with Network.connect
        """
        for phase_angle in [20, 75]:
            vc_designer = rf.Network(os.path.join(self.test_dir,
                                                  f"designer_variable_coupler_ideal_{phase_angle}deg.s4p"))
            vc_connect = self.variable_coupler_network_from_connect(phase_angle)
            assert_array_almost_equal(vc_designer.s, vc_connect.s, decimal=4)


class CircuitTestGraph(unittest.TestCase):
    """
    Test functionalities linked to graph method, used in particular for plotting
    """

    def setUp(self):
        """
        Dummy Circuit setup

        Setup a circuit which has various interconnections (2 or 3)
        """
        self.freq = rf.Frequency(start=1, stop=2, npoints=101, unit='GHz')

        # dummy components
        self.R = 100
        self.line_resistor = rf.media.DefinedGammaZ0(frequency=self.freq, z0=self.R)
        resistor1 = self.line_resistor.resistor(self.R, name='resistor1')
        resistor2 = self.line_resistor.resistor(self.R, name='resistor2')
        resistor3 = self.line_resistor.resistor(self.R, name='resistor3')
        port1 = rf.Circuit.Port(self.freq, name='port1')

        # Connection setup
        self.connections = [
                   [(port1, 0), (resistor1, 0), (resistor3, 0)],
                   [(resistor1, 1), (resistor2, 0)],
                   [(resistor2, 1), (resistor3, 1)]
                ]

        self.C = rf.Circuit(self.connections)

    def test_intersection_dict(self):
        inter_dict = self.C.intersections_dict
        # should have 3 intersections
        self.assertTrue(len(inter_dict) == 3)
        # All intersections should have at least 2 edges
        for it in inter_dict.items():
            k, cnx = it
            self.assertTrue(len(cnx) >= 2)

    def test_edge_labels(self):
        edge_labels = self.C.edge_labels
        self.assertTrue(len(edge_labels) == 7)


class CircuitTestComplexCharacteristicImpedance(unittest.TestCase):
    """
    Test creating circuits with real and complex port charac.impedances
    """
    def setUp(self):
        self.f0 = rf.Frequency(75.8, npoints=1, unit='GHz')
        # initial s-param values of A 2 ports network
        self.s0 = np.array([  # dummy values
            [-0.1000 -0.2000j, -0.3000 +0.4000j],
            [-0.3000 +0.4000j, 0.5000 -0.6000j]]).reshape(-1,2,2)

        # build initial network (under z0=50)
        self.ntw0 = rf.Network(frequency=self.f0, s=self.s0, z0=50, name='dut')

        # complex characteristic impedance to renormalize to
        self.zdut = 100 + 10j

        # reference solutions obtained from ANSYS Circuit or ADS (same res)
        # when z0=[50, zdut]
        self.z_ref = np.array([  # not affected by z0
            [18.0000 -16.0000j, 20.0000 +40.0000j],
            [20.0000 +40.0000j, 10.0000 -80.0000j]]).reshape(-1,2,2)

        self.y_ref = np.array([  # not affected by z0
            [0.0251 +0.0023j, 0.0123 -0.0066j],
            [0.0123 -0.0066j, 0.0052 +0.0055j]]).reshape(-1,2,2)

        self.s_ref = np.array([  # renormalized s (power-waves)
            [-0.1374 -0.2957j, -0.1995 +0.5340j],
            [-0.1995 +0.5340j, -0.0464 -0.7006j]]).reshape(-1,2,2)

        # Creating equivalent reference circuit
        port1 = rf.Circuit.Port(self.f0, z0=50, name='port1')
        port2 = rf.Circuit.Port(self.f0, z0=50, name='port2')
        ntw0 = self.ntw0

        cnx = [  # z0=[50,50]
            [(port1, 0), (ntw0, 0)],
            [(ntw0, 1), (port2, 0)]
            ]
        self.cir = rf.Circuit(cnx)

        # Creating equivalent circuit with z0 real
        port2_real = rf.Circuit.Port(self.f0, z0=100, name='port2')
        cnx_real = [  # z0=[50,100]
            [(port1, 0), (ntw0, 0)],
            [(ntw0, 1), (port2_real, 0)]
            ]
        self.cir_real = rf.Circuit(cnx_real)

        # Creating equivalent circuit with z0 complex
        port2_complex = rf.Circuit.Port(self.f0, z0=self.zdut, name='port2')
        cnx_complex = [  # z0=[50,zdut]
            [(port1, 0), (ntw0, 0)],
            [(ntw0, 1), (port2_complex, 0)]
            ]
        self.cir_complex = rf.Circuit(cnx_complex)

        # references for each s-param definition
        self.s_legacy = rf.renormalize_s(self.s0, [50,50], [50,self.zdut],
                                         s_def='traveling')
        self.s_power = rf.renormalize_s(self.s0, [50,50], [50,self.zdut],
                                         s_def='power')
        self.s_pseudo = rf.renormalize_s(self.s0, [50,50], [50,self.zdut],
                                         s_def='pseudo')
        # for real values, should be = whatever s-param definition
        self.s_real = rf.renormalize_s(self.s0, [50,50], [50,100])

    def test_verify_reference(self):
        ' Check that the reference results comes from power-waves definition'
        np.testing.assert_allclose(self.s_ref, self.s_power, atol=1e-4)

    def test_reference_s(self):
        ' Check reference z0 circuit, just in case '
        np.testing.assert_allclose(self.cir.network.s, self.s0, atol=1e-4)

    def test_real_z0_s(self):
        ' Check real z0 circuit '
        np.testing.assert_allclose(self.cir_real.network.s, self.s_real, atol=1e-4)

    def test_real_z_params(self):
        ' Check Z-parameters match'
        np.testing.assert_allclose(self.cir.network.z, self.cir_real.network.z, atol=1e-4)

    def test_complex_z_params(self):
        ' Check Z-parameters match'
        np.testing.assert_allclose(self.cir_complex.network.z, self.cir_real.network.z, atol=1e-4)

    @unittest.expectedFailure
    def test_complexz0_s_vs_legacy(self):
        ' Check complex z0 circuit vs legacy renormalization '
        np.testing.assert_allclose(self.cir_complex.network.s, self.s_legacy, atol=1e-4)

    def test_complexz0_s_vs_powerwaves(self):
        ' Check complex z0 circuit vs power-waves renormalization '
        np.testing.assert_allclose(self.cir_complex.network.s, self.s_ref, atol=1e-4)
        np.testing.assert_allclose(self.cir_complex.network.s, self.s_power, atol=1e-4)

    @unittest.expectedFailure
    def test_complexz0_s_vs_pseudo(self):
        ' Check complex z0 circuit vs pseudo-waves renormalization '
        np.testing.assert_allclose(self.cir_complex.network.s, self.s_pseudo, atol=1e-4)

class CircuitTestVoltagesCurrents(unittest.TestCase):
    def setUp(self):
        rng = np.random.default_rng()
        # setup a test transmission line randomly excited
        self.P_f = rng.random()  # forward power in Watt
        self.phase_f = rng.random()  # forward phase in rad
        self.Z = rng.random()  # source internal impedance, line characteristic impedance and load impedance
        self.L = rng.random()  # line length in [m]
        self.L2 = rng.random()  # line length in [m]
        self.L3 = rng.random()  # line length in [m]
        self.freq = rf.Frequency(1, 10, 10, unit='GHz')
        self.line_media = rf.media.DefinedGammaZ0(self.freq, z0=self.Z)  # lossless line medium
        self.line = self.line_media.line(d=self.L, unit='m', name='line')  # transmission line Network
        self.line2 = self.line_media.line(d=self.L2, unit='m', name='line2')  # transmission line Network
        self.line3 = self.line_media.line(d=self.L3, unit='m', name='line3')  # transmission line Network
        self.tee = self.line_media.tee(name='tee')
        self.resistor = self.line_media.resistor(50, name='resistor')

        # forward voltages and currents at the input of the test line
        self.V_in = np.sqrt(2*self.Z*self.P_f)*np.exp(1j*self.phase_f)
        self.I_in = np.sqrt(2*self.P_f/self.Z)*np.exp(1j*self.phase_f)
        # forward voltages and currents at the output of the test line
        theta = rf.theta(self.line_media.gamma, self.freq.f, self.L)  # electrical length
        self.V_out, self.I_out = rf.tlineFunctions.voltage_current_propagation(self.V_in, self.I_in, self.Z, theta)

        # Equivalent model with Circuit
        self.port1 = rf.Circuit.Port(frequency=self.freq, name='port1', z0=self.Z)
        self.port2 = rf.Circuit.Port(frequency=self.freq, name='port2', z0=self.Z)
        cnx = [
            [(self.port1, 0), (self.line, 0)],
            [(self.port2, 0), (self.line, 1)]
        ]
        self.crt = rf.Circuit(cnx)
        # power and phase arrays for Circuit.voltages() and currents()
        self.power = [self.P_f, 0]
        self.phase = [self.phase_f, 0]

    def test_tline_voltages(self):
        ' Test voltages for a simple transmission line '
        V_ports = self.crt.voltages_external(self.power, self.phase)

        np.testing.assert_allclose(self.V_in, V_ports[:,0])
        np.testing.assert_allclose(self.V_out, V_ports[:,1])

    def test_tline_currents(self):
        ' Test currents for a simple transmission line '
        I_ports = self.crt.currents_external(self.power, self.phase)

        np.testing.assert_allclose(self.I_in, I_ports[:,0])
        # output current is * -1 as Circuit definition is opposite
        # (toward the Circuit's Port)
        np.testing.assert_allclose(self.I_out, -1*I_ports[:,1])

    def test_multiple_ports(self):
        ' Test voltages and currents for a connection with multiple ports '
        # Create a circuit build with tee
        cnx_with_tee = [
            [(self.port1, 0), (self.line, 0)],
            [(self.line, 1), (self.tee, 0)],
            [(self.line2, 0), (self.tee, 1)],
            [(self.line3, 0), (self.tee, 2)],
            [(self.line2, 1), (self.port2, 0)],
            [(self.line3, 1), (self.resistor, 0)],
        ]

        # Create a circuit build without tee
        cnx = [
            [(self.port1, 0), (self.line, 0)],
            [(self.line, 1), (self.line2, 0), (self.line3, 0)],
            [(self.line2, 1), (self.port2, 0)],
            [(self.line3, 1), (self.resistor, 0)],
        ]

        # Create the two circuits
        crt_with_tee = rf.Circuit(cnx_with_tee)
        crt = rf.Circuit(cnx)

        # Get voltages and currents for both circuits
        V_with_tee = crt_with_tee.voltages(self.power, self.phase)
        I_with_tee = crt_with_tee.currents(self.power, self.phase)

        V = crt.voltages(self.power, self.phase)
        I = crt.currents(self.power, self.phase)

        # Get the reordering of the ports in the two circuits
        port_order_with_tee = (0, 1, 2, 4, 6, 8, 9, 10, 11)
        port_order = (0, 1, 2, 3, 4, 5, 6, 7, 8)

        for v, v_with_tee in zip((V, I), (V_with_tee, I_with_tee)):
            v_with_tee_reordered = np.array(
                [v_with_tee[:, i] for i in port_order_with_tee]
            )
            v_reordered = np.array([v[:, i] for i in port_order])

            np.testing.assert_allclose(v_with_tee_reordered, v_reordered)

    def test_tline_with_different_impedance(self):
        ' Test voltages and currents for a simple transmission line with different impedances '
        line = self.line.copy()
        line.renormalize(z_new=1./self.Z)

        # Equivalent model with Circuit
        cnx = [
            [(self.port1, 0), (line, 0)],
            [(self.port2, 0), (line, 1)]
        ]
        crt = rf.Circuit(cnx)

        V_ports_uni_z = crt.voltages(self.power, self.phase)
        I_ports_uni_z = crt.currents(self.power, self.phase)

        V_ports_dif_z = self.crt.voltages(self.power, self.phase)
        I_ports_dif_z = self.crt.currents(self.power, self.phase)

        np.testing.assert_allclose(I_ports_uni_z, I_ports_dif_z)
        np.testing.assert_allclose(V_ports_uni_z, V_ports_dif_z)

class CircuitTestVoltagesNonReciprocal(unittest.TestCase):
    def test_isolator(self):
        # Isolator that passes from port 1 to 2
        freq = rf.Frequency.from_f([1], unit='GHz')
        s = np.array([[[0, 0], [1, 0]]])
        network = rf.Network(frequency=freq, s=s, name='isolator')
        port1 = rf.Circuit.Port(frequency=freq, name='port1', z0=50)
        port2 = rf.Circuit.Port(frequency=freq, name='port2', z0=50)
        cnx = [
            [(network,0),(port1,0)],
            [(network,1),(port2,0)]
        ]
        crt = rf.Circuit(cnx)
        np.testing.assert_allclose(crt.s_external, s)

        power = [1,0] # 1 Watt at port 1
        phase = [0,0]
        V_at_ports = crt.voltages_external(power, phase)
        I_at_ports = crt.currents_external(power, phase)
        np.testing.assert_allclose(V_at_ports, [[10+0j, 10+0j]])
        # Positive current entering into port
        np.testing.assert_allclose(I_at_ports, [[0.2+0j, -0.2+0j]])

        power = [0,1] # 1 Watt at port 2
        V_at_ports = crt.voltages_external(power, phase)
        I_at_ports = crt.currents_external(power, phase)
        np.testing.assert_allclose(V_at_ports, [[0+0j, 10+0j]])
        np.testing.assert_allclose(I_at_ports, [[0+0j, 0.2+0j]])

    def test_isolator_reverse(self):
        # Isolator that passes from port 2 to 1
        freq = rf.Frequency.from_f([1], unit='GHz')
        s = np.array([[[0, 1], [0, 0]]])
        network = rf.Network(frequency=freq, s=s, name='isolator')
        port1 = rf.Circuit.Port(frequency=freq, name='port1', z0=50)
        port2 = rf.Circuit.Port(frequency=freq, name='port2', z0=50)
        cnx = [
            [(network,0),(port1,0)],
            [(network,1),(port2,0)]
        ]
        crt = rf.Circuit(cnx)
        np.testing.assert_allclose(crt.s_external, s)

        power = [1,0] # 1 Watt at port 1
        phase = [0,0]
        V_at_ports = crt.voltages_external(power, phase)
        I_at_ports = crt.currents_external(power, phase)
        np.testing.assert_allclose(V_at_ports, [[10+0j, 0+0j]])
        np.testing.assert_allclose(I_at_ports, [[0.2+0j, 0+0j]])

        power = [0,1] # 1 Watt at port 2
        V_at_ports = crt.voltages_external(power, phase)
        I_at_ports = crt.currents_external(power, phase)
        np.testing.assert_allclose(V_at_ports, [[10+0j, 10+0j]])
        np.testing.assert_allclose(I_at_ports, [[-0.2+0j, 0.2+0j]])

if __name__ == '__main__':
    unittest.main()



"""
.. module:: skrf.vi.vna
===============================================
vna (:mod:`skrf.vi.vna`)
===============================================


Provides classes to interact with Vector Network Analyzers (VNAs) from numerous
manufacturers.

VNA Classes
=======================================

.. autosummary::
    :nosignatures:
    :toctree: generated/

   hp.HP8510C
   keysight.FieldFox
   keysight.PNA
   nanovna.NanoVNAv2

The Base Class and Writing an Instrument Driver
===============================================

All VNAs are derived from the :class:`VNA` class. This class should never be
instantiated directly, but instead serves as a means to run certain setup code
depending on the capabilities of the instrument.

    * :ref:`SCPI`
    * :ref:`Channels`
    * :ref:`Property Creator`
    * :ref:`Validators`

When writing a new instrument driver, the following minimum questions should be
considered:
    * Does the instrument use SCPI?
    * Does the instrument support multiple channels?

SCPI
----

For those instruments that use SCPI, default universal SCPI methods are included
by default. However, for those instruments that do not support SPCI, do the
following:

.. code-block:: python

    class VNAWithoutSCPI(VNA):
        _scpi = False
        # ...

Channels
--------

Some instruments support channels and can have multiple measurements with
different frequency/power/if bandwidth/etc settings.

For instruments without channel support, the typical properties (freq_start,
freq_stop, etc) should be defined on the class. For instruments **with** channel
support, the class should include a subclass of :class:`Channel` that defines
properties specific to channel settings in the channel definition while other
instrument properties should be defined in the instrument class. To make this
more clear, consider the following example from the Keysight PNA implementation
(the property creator `VNA.Command` is explained later)

.. code-block:: python

    class PNA(VNA):
        class Channel(vna.Channel):
            freq_start = VNA.Command(
                get_cmd="SENS<self:cnum>:FREQ:STAR?",
                set_cmd="SENS<self:cnum>:FREQ:STAR <arg>",
                doc=\"""Start frequency [Hz]\""",
                validator=FreqValidator()
            )

        def __init__(self, address, backend='@py'):
            # ...

        nports = VNA.Command(
            get_cmd="SYST:CAP:HARD:PORT:COUN?",
            set_cmd=None,
            doc=\"""Number of ports\""",
            validator=IntValidator()
        )

Here, the start frequency is set **per channel** whereas the number of ports is
related to the instrument itself. Instruments with channel support should create a
single channel in `__init__()` using `create_channel`

Property Creator
----------------

Inspired by `PyMeasure <https://github.com/pymeasure/pymeasure>`_, skrf has a property
creator to simplify creating properties that are queried and set with commands.

.. automethod:: VNA.command

For `get_cmd` and `set_cmd`, you can substitute delimiters to be
replaced when the call is made. `self` is always passed, so you can
reference any part of "self" when making the call. Additionally, the
`set_cmd` receives an additional parameter `arg` which is the argument
passed to the setter. These can be used in the get/set strings by using
angle bracket delimiters and referring to the name. Here are some
examples:

Here, we are assuming we are writing a command for an instrument with
channels, and writing a command **for** the internal `Channel` class.
In `get_cmd`, `<self:cnum>` gets the `self.cnum` property of the Channel
class **at runtime**. In `set_cmd`, `<arg>` is replaced by the argument
passed to the setter.

.. code-block:: python

    freq_start = VNA.command(
        get_cmd="CALC<self:cnum>:FREQ:STAR?",
        set_cmd="CALC<self:cnum>:FREQ:STAR <arg>",
    )

And here's an example of calling this function and what the resultant
command strings would be after replacement.

.. code-block:: python

    _ = instr.ch1.freq_start
    # Sends the command CALC1:FREQ:STAR? to the instrument
    instr.ch1.freq_start = 100
    # Sends the command CALC1:FREQ:STAR 100 to the instrument


Validators
----------

Validators are (optionally, but almost always) passed to `VNA.command`. When a property
is get or set, the appropriate validate command is called to convert input to the proper
format expected by the instrument, or convert responses from the instrument to the
proper type.

The current validators are:
    * :class:`skrf.vi.validators.BooleanValidator`
    * :class:`skrf.vi.validators.DelimitedStrValidator`
    * :class:`skrf.vi.validators.DictValidator`
    * :class:`skrf.vi.validators.EnumValidator`
    * :class:`skrf.vi.validators.IntValidator`
    * :class:`skrf.vi.validators.FloatValidator`
    * :class:`skrf.vi.validators.FreqValidator`
    * :class:`skrf.vi.validators.SetValidator`

The documentation for each of these explains more about their functionality, but in essence
when writing a `VNA.command`, consider how the command must be formatted to be sent to the
instrument and what the expected response from the instrument will be and how that can be
transformed to something more useful than, say, a string.

Here's an example of using a validator:

.. code-block:: python

    freq_start = VNA.command(
        get_cmd="CALC:FREQ:STAR?",
        set_cmd="CALC:FREQ:STAR <arg>",
        validator=FreqValidator()
    )

    # ...

    # This will call FreqValidator.validate_input('100 kHz') to
    # transform the string '100 kHz' to '100000'. The full command
    # then becomes:
    # CALC:FREQ:STAR 100000
    instr.ch1.freq_start = '100 Hz'

    # This will send the command CALC:FREQ:STAR? to the instrument
    # then send the response from the instrument to
    # FreqValidator.validate_output() to attempt to transform, as an
    # example, the string '100000' to the int 100_000
    _ = instr.ch1.freq_start
"""
from .vna import VNA, Channel, ValuesFormat # isort: skip
from . import keysight, nanovna # isort: skip



r"""
Qfactor (:mod:`skrf.qfactor`)
========================================
Module for estimating the Quality (Q) factor(s) from S-parameters.

This class implements methods for determining *loaded* and *unloaded* Q-factor
from frequency-domain S-parameters, that can be applied to measurements
of transmission or reflection.

Documentation and implementation are adapted from [MAT58]_

Q-factor
--------
.. autosummary::
    :toctree: generated/
    :nosignatures:
    :recursive:

    Qfactor

.. currentmodule:: skrf.qfactor


Loaded and Unloaded Q-factor
----------------------------
The Quality factor (Q-factor) of a resonator is defined by [Pozar]_:

.. math::

    Q = \frac{2 \pi U}{\Delta U}


where :math:`U` is the average energy stored by the resonator and
:math:`\Delta U` is the decrease in the average stored energy per wave cycle
at the resonant frequency, that is, the average power loss.

The *loaded* Q-factor, :math:`Q_L`, describes energy dissipation within the
entire resonant system comprising of the resonator itself and the instrument
used for observing resonances. The term loading refers to the effect that the
external circuit has on measured quantities.

The external circuit consists of the measuring instrument and uncalibrated lines,
but not the couplings of microwave resonators. Loading by an instrument that
has 50 Ohm impedance, such as a VNA, causes :math:`Q_L` to be reduced substantially
if strong coupling is used.

For many applications the quantity that is desired is the *unloaded* Q-factor :math:`Q_0`,
which is determined by energy dissipation associated with the resonator only
and therefore gives the best description of the resonant mode(s).

In other words, :math:`Q_0` is the Q-factor of the uncoupled resonator. The value of
:math:`Q_0` can be estimated from measurements of :math:`Q_L`, but cannot be measured directly.
:math:`Q_0` is largely governed by ohmic loss arising from surface currents
in the metal conductors (walls and loop couplings), and from dielectric loss
in any insulating materials that may be present.


Relationships between Loaded and Unloaded Q-factors
---------------------------------------------------
Energy dissipation in the external circuit is characterised by the *external* Q-factor,
:math:`Q_e`. For both series and parallel equivalent circuits, the three
Q-factors are related by [Pozar]_:

.. math::

    \frac{1}{Q_L} = \frac{1}{Q_0} + \frac{1}{Q_e}


The coupling factor :math:`\beta` is defined for each port as:

.. math::

    \beta = \frac{Q_0}{Q_e}


Finding the unloaded Q from measured S-parameters consists in first finding
the coupling factor, then measure :math:`Q_L` from the 3 dB bandwidth
and using the relationships above.

Fortunately, scikit-rf implements methods for determining loaded and
unloaded Q-factors from frequency-domain S-parameters. The implemented methods
are described in detail in [MAT58]_, and can be applied to measurements of
transmission or reflection.

Q-factor determination through equivalent-circuit models
--------------------------------------------------------
Characterisation of resonances from measurements in the frequency-domain
can be achieved through equivalent-circuit models [MAT58]_. Resonators can be
modelled as an ideal RLC resonator connected to an external circuit,
incorporating elements to account for a lossy coupling and coupling reactances.

For high Q-factor resonators (in practice, :math:`Q_L` > 100), the S-parameter
response of a resonator measured in a calibrated system with reference planes
at the resonator couplings can be expressed like [MAT58]_, [Galwas]_ :

.. math::

    S = S_D + d \frac{e^{−2j\delta}}{1 + j Q_L t}


where :math:`S_D` is the detuned S-parameter measured at frequencies far above or below
resonance, :math:`d` is The diameter of the Q-circle, :math:`\delta` is a
real-valued constant that defines the orientation of the Q-circle, and :math:`t`
is the fractional offset frequency given by:

.. math::

    t = \frac{f}{f_L} - \frac{f_L}{f}  \approx 2 \frac{f − f_L}{f_L}


where :math:`f_L` is the loaded resonant frequency and :math:`f` the frequency
at which S is measured. This equation can be applied to measurements
by transmission (S21 or S12) or reflection (S11 or S22).

The S-parameters are fitted against a modified expression of the above equations
to deduce the resonant frequency, loaded and unloaded Q-factors and other properties.

References
----------
.. [MAT58] "Q-factor Measurement by using a Vector Network Analyser",
    A. P. Gregory, National Physical Laboratory Report MAT 58 (2021)
    https://eprintspublications.npl.co.uk/9304/

.. [Pozar] D. M. Pozar, Microwave engineering, 4th ed. J. Wiley, 2012.

.. [Galwas] B. A. Galwas, ‘Scattering Matrix Description of Microwave
   Resonators’, IEEE Trans. Microwave Theory Techn., vol. 31, no. 8,
   pp. 669–671, Aug. 1983, doi: 10.1109/TMTT.1983.1131566.

"""
from __future__ import annotations

from warnings import warn

import numpy as np

from .constants import NumberLike
from .frequency import Frequency
from .network import Network

# Available resonance types
RESONANCE_TYPES = ['reflection', 'reflection_method2',
                   'transmission', 'absorption']


class OptimizedResult(dict):
    """Represent Q-factor optimisation result.

    Attributes
    ----------
    Q_L : float
        Loaded Quality factor
    f_L : float
        Resonance frequency [Hz]
    success: bool
        Is the fit method has been successfully performed
    method: str
        Fitting method used.
    m1, m2, m3, ...: float
        Coefficients described in [MAT58]_
    number_iterations: int
        Number of iterations performed.

    Notes
    -----
    `OptimizedResult` may have additional attributes not listed here depending
    on the specific fitting method used. Since this class is essentially a
    subclass of dict with attribute accessors, one can see which
    attributes are available using the `OptimizedResult.keys` method.

    References
    ----------
    .. [MAT58] "Q-factor Measurement by using a Vector Network Analyser",
        A. P. Gregory, National Physical Laboratory Report MAT 58 (2021)
        https://eprintspublications.npl.co.uk/9304/

    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(name) from e

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in sorted(self.items())])
        else:
            return self.__class__.__name__ + "()"

    def __dir__(self):
        return list(self.keys())


class Qfactor:
    """
    Q-factor calculation class.

    This class implements methods for determining *loaded* and *unloaded* Q-factor
    from frequency-domain S-parameters, that can be applied to measurements
    of transmission or reflection.

    Parameters
    ----------
    ntwk : :class:`~skrf.network.Network` object
        A 1-port scikit-rf Network.
        If your device is a N-port, pass the desired sub-S-parameters to fit
        the data from, like `ntwk.s21`.
    res_type : str
        Specifies the resonance type: 'reflection', 'transmission',
        'reflection_method2' or 'absorption':
        'reflection' is generally suited for undercoupled resonators,
        while 'reflection_method2' is favoured for coupling with large loop.
    Q_L0 : float, optional. Default is None.
        Estimated loaded Q-factor, used to improve fitting.
    f_L0 : float, optional. Default is None.
        Estimated loaded resonant frequency, used to improve fitting [Hz]
        If None, automatically search for the min or max,
        depending on the resonance type defined by `res_type`.
    verbose : bool, optional. Default is False.
        Boolean flag controlling output of information to the console.

    Raises
    ------
    ValueError
        If the passed Network is not 1-port or if the resonance type is unknown.

    Notes
    -----
    Uncalibrated line should be de-embedded (if it has significant
    length) from the S-parameter data before calling the functions
    in this module to get best results.

    """

    def __init__(self,
                 ntwk: Network,
                 res_type: str,
                 Q_L0: None | float = None,
                 f_L0: None | float = None,
                 verbose: bool = False):
        """Q-factor initializer."""
        # check ntwk is a 1-port
        if ntwk.nports != 1:
            raise ValueError('The Network is not a 1-port Network.')
        if res_type not in RESONANCE_TYPES:
            raise ValueError(f'res_type must be in: {RESONANCE_TYPES}.')

        self.s = ntwk.s
        self.f = ntwk.f
        self.f_scaled = ntwk.frequency.f_scaled
        self.f_multiplier = ntwk.frequency.multiplier
        self.f_unit = ntwk.frequency.unit
        self._ntwk = ntwk
        self.res_type = res_type
        self.tol = 1.0e-5
        self.verbose = verbose
        self.fitted = False
        self.opt_res = None

        self.N = len(self.f)

        # step 1: initial_fit. Deduce premilinary values for Q_L and f_L.
        self._initial_fit(self.N, Q_L0, f_L0)

    def __str__(self) -> str:
        if self.fitted:
            status = f"fitted: f_L={float(self.f_L/self.f_multiplier):.3f}{self.f_unit}, Q_L={float(self.Q_L):.3f}"
        else:
            status = 'not fitted'

        _str = f"Q-factor of Network {self._ntwk.name}. ({status})"
        return _str

    def __repr__(self) -> str:
        return self.__str__()

    def fit(self,
            method: str = "NLQFIT6",
            loop_plan: str = 'fwfwc'
            ) -> OptimizedResult:
        """Fit Q-factor from S-parameter data.

        Fitting overwrites the parameters `Q_L` and `f_L`.

        Parameters
        ----------
        method : str, optional
            Fitting method : 'NLQFIT6' (default), 'NLQFIT7', 'NLQFIT8':

            - 'NLQFIT6': Least Square Minimum of Eq.21 [MAT58]_ with 6 coeffcients.
            - 'NLQFIT7': Least Square Minimum of Eq.26 [MAT58]_ with 7 coeffcients,
              including one that characterize the trans. line length.
            - 'NLQFIT8': Least Square Minimum of Eq.43 [MAT58]_ with 8 coeffcients,
              A model for frequency-dependent leakage.
        loop_plan : str, optional
            Defines order of steps used by the fitting process.
            The convergence algorithm uses a number of steps set by loop_plan,
            a string of characters as follows:

            - 'f': fit once without testing for convergence
            - 'c': repeated fit, iterating until convergence is obtained
            - 'w': re-calculate weighting factors on basis of previous fit
            - Initially the weighting factors are all unity.
            - The first character in `loop_plan` must not be 'w'.
            e.g.: 'fwfwc' (default).

        Returns
        -------
        result : :class:`~skrf.qfactor.OptimizedResult`

        References
        ----------
        .. [MAT58] "Q-factor Measurement by using a Vector Network Analyser",
            A. P. Gregory, National Physical Laboratory Report MAT 58 (2021)
            https://eprintspublications.npl.co.uk/9304/

        """

        for op in loop_plan:
            if op not in ['f', 'c', 'w']:
                raise ValueError("Unexpected character in loop_plan")
        if loop_plan[-1] == "w":
            raise ValueError("Last item in loop_plan must not be w (weight calculation)")
        if loop_plan[0] == "w":
            raise ValueError("First item in loop_plan must not be w (weight calculation)")
        if loop_plan[-1] != "c":
            warn("Last item in loop_plan is not c so convergence not tested!", stacklevel=2)

        self.method = method
        self.loop_plan = loop_plan

        # step 2: least square fitting
        if method == 'NLQFIT6':
            result = self._optimise_fit6(self.N)
        elif method == 'NLQFIT7':
            result = self._optimise_fit7(self.N)
        elif method == 'NLQFIT8':
            result = self._optimise_fit8(self.N)

        # overwrite results in self
        self.Q_L = result.Q_L
        self.f_L = result.f_L
        self.fitted = True
        self.opt_res = result

        if result.Q_L < 0:
            warn('Negative Q_L, fitting may be inaccurate.', stacklevel=2)

        return result

    @staticmethod
    def angular_weights(f: NumberLike,
                        f_L: NumberLike,
                        Q_L: NumberLike
                        ) -> NumberLike:
        r"""Diagonal elements W_i of weights matrix.

        .. math::

            W_i = \frac{1}{\left[ \frac{2 Q_L (f_i - f_L)}{f_L} \right]^2 + 1}


        The weights are needed when frequencies are equally-spaced
        (rather than points equally spaced around the Q-circle), and help
        reducing systematic error [MAT58]_.

        Parameters
        ----------
        f : np.ndarray
            Frequency values array.
        f_L : float
            Loaded resonant frequency.
        Q_L : float
            Loaded Q-factor.

        Returns
        -------
        W_i : np.ndarray
            Weighting factors in proportion to the rate of change of angle
            with frequency relative to the centre of the Q-circle.

        References
        ----------
        .. [MAT58] "Q-factor Measurement by using a Vector Network Analyser",
            A. P. Gregory, National Physical Laboratory Report MAT 58 (2021)
            https://eprintspublications.npl.co.uk/9304/
            section 2.4, eqn. (28).

        """
        ptmp = 2 * Q_L * (f - f_L) / f_L
        W_i = 1 / (ptmp ** 2 + 1)
        return W_i

    def _initial_fit(self,
                    N: int,
                    Q_L0: float | None = None,
                    f_L0: None | float = None
                    ):
        """Initial Linear least squares Q-factor fit.

        As this is not optimised in this function (use `fit`), the solution
        will only be approximate. This method is called during the
        initialization of the `Qfactor` class. Note that a reasonable estimate
        for the resonant frequency should be supplied is multiple resonances
        are present.

        Also calculate the internal parameters a, b and QL from ([MAT58]_, eqn. 17)

        Parameters
        ----------
        N : int
            Number of points.
        Q_L0 : float, optional. Default is None.
            Estimated loaded Q-factor (will be improved by fitting).
        f_L0 : float, optional. Default is None.
            Estimated loaded resonant frequency, used to improve fitting [Hz]

        Returns
        -------
        None

        References
        ----------
        .. [MAT58] "Q-factor Measurement by using a Vector Network Analyser",
            A. P. Gregory, National Physical Laboratory Report MAT 58 (2021)
            https://eprintspublications.npl.co.uk/9304/
            section 2.1, eqn. (17).

        """
        if f_L0 is None:
            # search for the initial value of the resonance frequency
            if self.res_type in ['reflection', 'reflection_method2', 'absorption']:
                # Find minimum in |S11|
                index_min = np.argmin(np.abs(self.s))
                f_L0 = self.f[index_min]
            else:
                # Find peak in |S21|
                index_max = np.argmax(np.abs(self.s))
                f_L0 = self.f[index_max]

        # Q_L0 : An order-of-magnitude estimate for Q_L-factor
        if Q_L0 is None:
            # The value 5.0 should work well
            # for initial and optimised fits (Section 2.6).
            mult = 5.0
            Q_L0 = mult * f_L0/(self.f[-1] - self.f[0])

        if self.verbose:
            print(f'Initial estimation: Q_L0={Q_L0}, f_L0={f_L0}')

        N2 = 2 * N
        M = np.zeros([N2, 5])
        G = np.zeros(N2)[:, np.newaxis]

        for i in range(N):
            i2 = i + N
            t = 2.0 * (self.f[i] / f_L0 - 1.0)
            y = 1.0 / complex(1.0, Q_L0 * t)
            v = t * y
            v1 = y * self.s[i, 0, 0]
            G[i] = v1.real
            G[i2] = v1.imag
            v2 = v1 * t
            M[i, :] = np.array([v.real, -v.imag, y.real, -y.imag, v2.imag])
            M[i2, :] = np.array([v.imag, v.real, y.imag, y.real, -v2.real])

        T = M.transpose()  # unweighted
        C = T @ M
        q = T @ G
        sv = np.linalg.solve(C, q)
        a_re, a_im, b_re, b_im, Q_L = sv

        self._a = (a_re + 1j*a_im)[0]
        self._b = (b_re + 1j*b_im)[0]
        self.Q_L = Q_L[0]
        self.f_L = f_L0

        if self.verbose:
            print(f'Preliminary estimation: Q_L={self.Q_L}, f_L={self.f_L}')


    def _optimise_fit6(self, N: int):
        """Iterative non-linear fit, NLQFIT6 Step (2).

        Optimised fit of Q-factor (Q_L) and resonant frequency (f_L)
        by the gradient-descent method [MAT58]_.

        Uses the results of the initial fit as the starting
        values for the iteration.

        Parameters
        ----------
        N : int
            Number of points.

        Returns
        -------
        res : OptimizedResult
            Fitted values.

        References
        ----------
        .. [MAT58] MAT 58, section 2.2, eqn. (22).

        """
        N2 = N * 2
        iterations = 0
        PV = np.ones(N)  # default weights vector
        PV2 = np.ones(N2)

        m1 = self._a.imag / self.Q_L  # a''/QL
        m2 = -self._a.real / self.Q_L
        m3 = self._b.real - m1
        m4 = self._b.imag - m2
        m5 = self.Q_L
        Flwst = self.f[0]  # lowest freq. is a convenient normalisation factor.
        m6 = Flwst * m5 / self.f_L
        last_op = "n"
        weighting_ratio = None
        number_iterations = 0

        ## Loop through all of the operations specified in loop_plan
        for op in self.loop_plan:
            if op == "w":
                PV = self.angular_weights(self.f, Flwst * float(m5) / float(m6), float(m5))
                # PV = self.angular_weights(m5)
                weighting_ratio = max(PV) / min(PV)
                PV2 = np.concatenate((PV, PV))
                if self.verbose:
                    print("Op w, Calculate weights")
                last_op = "n"
            elif op == "c":
                seek_convergence = True
            elif op == "f":
                seek_convergence = False

            TerminationConditionMet = False
            RMS_Error = None
            while not (TerminationConditionMet):
                number_iterations += 1
                M = np.zeros([N2, 6])  # X is the transpose of M
                G = np.zeros(N2)[:, np.newaxis]
                c1 = complex(-m4, m3)
                c2 = complex(m1, m2)
                c3 = complex(m3, m4)
                for i in range(N):
                    i2 = i + N
                    y = 1.0 / complex(1.0, 2 * (m6 * self.f[i] / Flwst - m5))
                    u = c1 * y * y * 2
                    u2 = -u * self.f[i] / Flwst
                    M[i, :] = 1.0, 0.0, y.real, -y.imag, u.real, u2.real
                    M[i2, :] = 0.0, 1.0, y.imag, y.real, u.imag, u2.imag
                    v = c2 + c3 * y
                    r = self.s[i] - v  # residual
                    G[i] = r.real
                    G[i2] = r.imag
                X = M.transpose()
                T = np.multiply(X, PV2)
                C = np.dot(T, M)
                q = np.dot(T, G)
                dm = np.linalg.solve(C, q)[:,0]
                m1 += dm[0]
                m2 += dm[1]
                m3 += dm[2]
                m4 += dm[3]
                m5 += dm[4]
                m6 += dm[5]
                del G, X, T, C, dm
                iterations = iterations + 1
                if RMS_Error is not None:
                    Last_RMS_Error = RMS_Error
                else:
                    Last_RMS_Error = None
                SumNum = 0.0
                SumDen = 0.0
                for i in range(N):
                    den = complex(1.0, 2 * (m6 * self.f[i] / Flwst - m5))
                    ip = PV[i]
                    E = self.s[i] - complex(m1, m2) - complex(m3, m4) / den
                    SumNum = SumNum + ip * (E.real * E.real + E.imag * E.imag)
                    SumDen = SumDen + ip
                RMS_Error = np.sqrt(SumNum / SumDen)
                if self.verbose:
                    if last_op == "c":
                        print(f"Iteration {iterations}, RMS Error: {RMS_Error}")
                    else:
                        print(f"op {op}, Iteration {iterations}, RMS Error: {RMS_Error}")
                last_op = op

                if seek_convergence:
                    if Last_RMS_Error is not None:
                        delta_S = abs(RMS_Error - Last_RMS_Error)
                        TerminationConditionMet = delta_S < self.tol
                else:
                    TerminationConditionMet = True
            # After last operation, we end up here ...
            if self.verbose:
                print("Optimization done.")


        return OptimizedResult({
            'success': TerminationConditionMet,
            'm1': m1, 'm2': m2, 'm3': m3, 'm4': m4,
            'Q_L': m5,
            'f_L': m5 * Flwst / m6,
            'weighting_ratio': weighting_ratio,
            'number_iterations': number_iterations,
            'RMS_Error': RMS_Error,
            'method': self.method,
            })

    def _optimise_fit7(self, N):
        """Iterative non-linear fit, NLQFIT7 Step (2).

        Optimised fit of Q-factor (QL) and resonant frequency (FL)
        by the gradient-descent method.

        Uses the results of the initial fit (sv) as the starting
        values for the iteration.

        Parameters
        ----------
        F : np.ndarray
            Frequency points
        S : np.ndarray
            Complex data S-parameter to be fitted.
        N : int
            Number of points.
        Fseed : float
            Estimated resonant frequency.
        sv : list
            Initial solution (numpy vector or a list) found with initial_fit.
        loop_plan : str
            Characters which defines order of steps used by the fitting process
            e.g. 'fwfwc':

            - 'f': fit once without testing for convergence.
            - 'c': repeated fit, iterating until convergence is obtained.
            - 'w': re-calculate weighting factors on basis of previous fit.
        Tol : float
            Criterion for the convergence test.
            Recommend using 1.0E-5 for reflection or max(abs(S))*1.0E-5
            for transmission.
        quiet : bool
            Boolean flag controlling output of information to the console.

        Returns
        -------
        list
            list of fitted parameters: [m1, m2, m3, m4, m5, m5 * Flwst / m6, m7 / Flwst]
        weighting_ratio : float
        number_iterations : int
        RMS_Error : float

        References
        ----------
        .. [MAT58] MAT 58, section 2.3, eqn. (26).

        """
        N2 = N * 2
        iterations = 0
        PV = np.ones(N)  # default weights vector
        PV2 = np.ones(N2)

        m1 = self._a.imag / self.Q_L  # a''/QL
        m2 = -self._a.real / self.Q_L
        m3 = self._b.real - m1
        m4 = self._b.imag - m2
        m5 = self.Q_L
        Flwst = self.f[0]  # lowest freq. is a convenient normalisation factor.
        m6 = Flwst * self.Q_L / self.f_L
        m7 = 0.0
        last_op = "n"
        weighting_ratio = None
        number_iterations = 0

        ## Loop through all of the operations specified in loop_plan
        for op in self.loop_plan:

            if op == "w":
                PV = self.angular_weights(self.f, Flwst * m5 / m6, m5)
                weighting_ratio = max(PV) / min(PV)
                PV2 = np.concatenate((PV, PV))
                if self.verbose:
                    print("Op w, Calculate weights")
                last_op = "n"
                continue
            if op == "c":
                seek_convergence = True
            elif op == "f":
                seek_convergence = False
            else:
                assert 0, "Unexpected character in loop_plan"

            TerminationConditionMet = False
            RMS_Error = None
            while not (TerminationConditionMet):
                number_iterations += 1
                M = np.zeros([N2, 7])
                G = np.zeros(N2)[:, np.newaxis]
                c1 = complex(-m4, m3)
                c2 = complex(m1, m2)
                c3 = complex(m3, m4)
                for i in range(N):
                    i2 = i + N
                    y = 1.0 / complex(1.0, 2 * (m6 * self.f[i] / Flwst - m5))
                    fdn = self.f[i] / Flwst - m5 / m6
                    pj = complex(0.0, m7 * fdn)
                    expm7 = np.exp(pj)
                    ym = y * expm7
                    u = c1 * y * ym * 2
                    u2 = -u * self.f[i] / Flwst
                    v = (c2 + y * c3) * expm7
                    u3 = v * fdn
                    M[i, :] = np.array(
                        [expm7.real,
                        -expm7.imag,
                        ym.real,
                        -ym.imag,
                        u.real,
                        u2.real,
                        -u3.imag]
                    )
                    M[i2, :] = np.array(
                        [expm7.imag,
                        expm7.real,
                        ym.imag,
                        ym.real,
                        u.imag,
                        u2.imag,
                        u3.real]
                    )
                    r = self.s[i] - v  # residual
                    G[i] = r.real
                    G[i2] = r.imag
                X = M.transpose()
                T = np.multiply(X, PV2)
                C = np.dot(T, M)
                q = np.dot(T, G)
                dm = np.linalg.solve(C, q)[:,0]
                m1 += dm[0]
                m2 += dm[1]
                m3 += dm[2]
                m4 += dm[3]
                m5 += dm[4]
                m6 += dm[5]
                m7 += dm[6]
                del G, X, T, C, dm
                iterations = iterations + 1
                if RMS_Error is not None:
                    Last_RMS_Error = RMS_Error
                else:
                    Last_RMS_Error = None

                SumNum = 0.0
                SumDen = 0.0
                for i in range(N):
                    fdn = self.f[i] / Flwst - m5 / m6
                    den = complex(1.0, 2 * (m6 * self.f[i] / Flwst - m5))
                    pj = complex(0.0, m7 * fdn)
                    E = self.s[i] - (c2 + c3 / den) * np.exp(pj)
                    ip = PV[i]
                    SumNum = SumNum + ip * (E.real * E.real + E.imag * E.imag)
                    SumDen = SumDen + ip
                RMS_Error = np.sqrt(SumNum / SumDen)
                if self.verbose:
                    if last_op == "c":
                        print(f"Iteration {iterations}, RMS Error: {RMS_Error}")
                    else:
                        print(f"op {op}, Iteration {iterations}, RMS Error: {RMS_Error}")

                last_op = op

                if seek_convergence:
                    if Last_RMS_Error is not None:
                        delta_S = abs(RMS_Error - Last_RMS_Error)
                        TerminationConditionMet = delta_S < self.tol
                else:
                    TerminationConditionMet = True
            # After last operation, we end up here ...
            if self.verbose:
                print("Optimization done.")

        return OptimizedResult({
            'success': TerminationConditionMet,
            'm1': m1, 'm2': m2, 'm3': m3, 'm4': m4,
            'Q_L': m5,
            'f_L': m5 * Flwst / m6,
            'm7a' : m7 / Flwst,
            'weighting_ratio': weighting_ratio,
            'number_iterations': number_iterations,
            'RMS_Error': RMS_Error,
            'method': self.method,
            })

    def _optimise_fit8(self, N):
        """Iterative non-linear fit, NLQFIT8 Step (2).

        Optimised fit of Q-factor (QL) and resonant frequency (FL)
        by the gradient-descent method.

        Uses the results of the initial fit (sv) as the starting
        values for the iteration.

        Parameters
        ----------
        F : np.ndarray
            Frequency points
        S : np.ndarray
            Complex data S-parameter to be fitted.
        N : int
            Number of points.
        Fseed : float
            Estimated resonant frequency.
        sv : list
            Initial solution (numpy vector or a list) found with initial_fit.
        loop_plan : str
            Characters which defines order of steps used by the fitting process
            e.g. 'fwfwc':
                'f' - fit once without testing for convergence.
                'c' - repeated fit, iterating until convergence is obtained.
                'w' - re-calculate weighting factors on basis of previous fit.
        Tol : float
            Criterion for the convergence test.
                    Recommend using 1.0E-5 for reflection or max(abs(S))*1.0E-5
                    for transmission.
        quiet : bool
            Boolean flag controlling output of information to the console.

        Returns
        -------
        list
            list of fitted parameters: [m1, m2, m3, m4, m5, m5 * Flwst / m6, m7 / Flwst]
        weighting_ratio : float
        number_iterations : int
        RMS_Error : float

        References
        ----------
        .. [MAT58] MAT 58, sec 4.5, eqn. (43).

        """
        N2 = N * 2
        iterations = 0
        PV = np.ones(N)  # default weights vector
        PV2 = np.ones(N2)

        m1 = self._a.imag / self.Q_L  # a''/QL
        m2 = -self._a.real / self.Q_L
        m3 = self._b.real - m1
        m4 = self._b.imag - m2
        m5 = self.Q_L
        Flwst = self.f[0]  # lowest freq. is a convenient normalisation factor.
        m6 = Flwst * self.Q_L / self.f_L
        m8 = 0.0
        m9 = 0.0
        last_op = "n"
        weighting_ratio = None
        number_iterations = 0

        ## Loop through all of the operations specified in loop_plan
        for op in self.loop_plan:

            if op == "w":  # Fr                       QL
                PV = self.angular_weights(self.f, Flwst * float(m5) / float(m6), float(m5))
                weighting_ratio = max(PV) / min(PV)
                PV2 = np.concatenate((PV, PV))
                if self.verbose:
                    print("Op w, Calculate weights")
                last_op = "n"
                continue
            if op == "c":
                seek_convergence = True
            elif op == "f":
                seek_convergence = False
            else:
                assert 0, "Unexpected character in loop_plan"

            TerminationConditionMet = False
            RMS_Error = None
            while not (TerminationConditionMet):
                number_iterations += 1
                M = np.zeros([N2, 8])
                G = np.zeros(N2)[:, np.newaxis]
                c1 = complex(-m4, m3)
                c2 = complex(m1, m2)
                c3 = complex(m3, m4)
                for i in range(N):
                    i2 = i + N
                    y = 1.0 / complex(1.0, 2 * (m6 * self.f[i] / Flwst - m5))
                    u = c1 * y * y * 2
                    u2 = -u * self.f[i] / Flwst
                    FL = Flwst * m5 / m6
                    t = 2 * (self.f[i] - FL) / FL
                    M[i, :] = np.array([1.0, 0.0, y.real, -y.imag, u.real, u2.real, t, 0.0])
                    M[i2, :] = np.array([0.0, 1.0, y.imag, y.real, u.imag, u2.imag, 0.0, t])
                    v = c2 + c3 * y + (m8 + 1j * m9) * t
                    r = self.s[i] - v  # residual
                    G[i] = r.real
                    G[i2] = r.imag
                X = M.transpose()
                T = np.multiply(X, PV2)
                C = np.dot(T, M)
                q = np.dot(T, G)
                dm = np.linalg.solve(C, q)[:,0]
                m1 += dm[0]
                m2 += dm[1]
                m3 += dm[2]
                m4 += dm[3]
                m5 += dm[4]
                m6 += dm[5]
                m8 += dm[6]
                m9 += dm[7]
                del G, X, T, C, dm
                iterations = iterations + 1
                if RMS_Error is not None:
                    Last_RMS_Error = RMS_Error
                else:
                    Last_RMS_Error = None

                SumNum = 0.0
                SumDen = 0.0
                for i in range(N):
                    den = complex(1.0, 2 * (m6 * self.f[i] / Flwst - m5))
                    FL = Flwst * m5 / m6
                    t = 2 * (self.f[i] - FL) / FL
                    ip = PV[i]
                    E = (
                        self.s[i]
                        - complex(m1, m2)
                        - complex(m8, m9) * t
                        - complex(m3, m4) / den
                    )
                    SumNum = SumNum + ip * (E.real * E.real + E.imag * E.imag)
                    SumDen = SumDen + ip
                RMS_Error = np.sqrt(SumNum / SumDen)
                if self.verbose:
                    if last_op == "c":
                        print(f"Iteration {iterations}, RMS Error: {RMS_Error}")
                    else:
                        print(f"{op}, Iteration {iterations}, RMS Error: {RMS_Error}")

                last_op = op

                if seek_convergence:
                    if Last_RMS_Error is not None:
                        delta_S = abs(RMS_Error - Last_RMS_Error)
                        TerminationConditionMet = delta_S < self.tol
                else:
                    TerminationConditionMet = True
            # After last operation, we end up here ...
            if self.verbose:
                print("Optimization done.")

        return OptimizedResult({
            'success': TerminationConditionMet,
            'm1': m1, 'm2': m2, 'm3': m3, 'm4': m4,
            'Q_L': m5,
            'f_L': m5 * Flwst / m6,
            'weighting_ratio': weighting_ratio,
            'number_iterations': number_iterations,
            'RMS_Error': RMS_Error,
            'method': self.method,
            })

    def Q_circle(self,
                 opt_res: None | OptimizedResult = None,
                 A: None | float = None
                 ) -> list:
        r"""Q-circle diameter.

        The diameter of the Q-circle (as displayed in a VNA) provides a visual
        indication of whether the coupling is strong or weak [MAT58]_.

        Parameters
        ----------
       opt_res : None or :class:`~skrf.qfactor.OptimizedResult`. Default is None.
           Solution produced by the :meth:`~skrf.qfactor.Qfactor.fit` method.
           If None, uses the solution previously calculated, if performed.
        A : None of float. Default is None.
            Scaling factor as defined in MAT 58 [MAT58]_.
            For `reflection` resonance type, can be set as None
            to use the magnitude of the fitted detuned reflection coefficient S_V

        Returns
        -------
        diam : float
            Q-circle diameter d.
        S_V : complex
            Off-resonance Reflection Coefficient.
        S_T : complex
            Tuned Reflection Coefficient.

        References
        ----------
        .. [MAT58] "Q-factor Measurement by using a Vector Network Analyser",
            A. P. Gregory, National Physical Laboratory Report MAT 58 (2021)
            https://eprintspublications.npl.co.uk/9304/,
            section 2.5, eqn. (31).

        """
        # if no solution passed, use internal solution if exist
        if opt_res is None:
            if self.opt_res is None:
                raise ValueError('No solution found or passed.')
            else:
                opt_res = self.opt_res

        # m1 : real part of cal_gamma_V
        # m2 : imag part of cal_gamma_V
        # m3 : real part of b + j a/Q_L
        # m4 : imag part of b + j a/Q_L
        m1, m2, m3, m4 = (opt_res[key] for key in ['m1', 'm2', 'm3', 'm4'])

        if A is None:
            A = 1.0 / abs(complex(m1, m2))  # scale to S_V
        elif not isinstance(A, (int, float)):
            raise ValueError("A should be a float or None")

        aqratio = complex(m1, m2)
        b = complex(m1 + m3, m2 + m4)
        diam = abs(b - aqratio) * A
        S_V = complex(m1, m2) * A
        S_T = b * A
        return diam, S_V, S_T


    def Q_unloaded(self,
                   opt_res: None| OptimizedResult = None,
                   A: None | float = None
                   ) -> float:
        """Unloaded Q-factor Q0.

        The value of the unloaded Q-factor Q0 cannot be measured directly but
        can be estimated from the measurement of the loaded Q-factor Q_L [MAT58]_ .

        Parameters
        ----------
       opt_res : None or :class:`~skrf.qfactor.OptimizedResult`. Default is None.
           Solution produced by the :meth:`~skrf.qfactor.Qfactor.fit` method.
           If None, uses the solution previously calculated, if performed.
        A : float or None. Default is None.
            Scaling factor as defined in MAT 58 [MAT58]_.
            For `reflection` resonance type, can be set as None
            to use the magnitude of the fitted detuned reflection coefficient S_V

        Returns
        -------
        Q0 : float
            Unloaded Q-factor.

        References
        ----------
        .. [MAT58] "Q-factor Measurement by using a Vector Network Analyser",
            A. P. Gregory, National Physical Laboratory Report MAT 58 (2021)
            https://eprintspublications.npl.co.uk/9304/

        """
        # if no solution passed, use internal solution if exist
        if opt_res is None:
            if self.opt_res is None:
                raise ValueError('No solution found or passed.')
            else:
                opt_res = self.opt_res

        if A is None:
            auto_flag = True
        elif isinstance(A, (int, float)):
            auto_flag = False
        else:
            raise ValueError("Illegal Scaling factor; should be a float or  None")

        m1, m2, m3, m4, m5 = (opt_res[key] for key in ['m1', 'm2', 'm3', 'm4', 'Q_L'])

        if self.res_type == "transmission":
            if auto_flag:
                raise ValueError('Scaling factor must be defined for transmission case')
            cal_diam, cal_gamma_V, cal_gamma_T = self.Q_circle(opt_res, A)
            if cal_diam == 1.0:
                raise ZeroDivisionError("Divide by zero forestalled in calculation of Q0")
            Q0 = m5 / (1.0 - cal_diam)

        elif self.res_type == "reflection":
            if auto_flag:
                if self.verbose:
                    print('A is undefined: using fitted data to estimate it')
                A = 1.0 / abs(complex(m1, m2))  # scale to S_V if A not defined
            cal_diam, S_V, S_T = self.Q_circle(opt_res, A)
            cal_touching_circle_diam = 2.0
            if self.verbose:
                print(f"Q-circle diam = {cal_diam}, touching_circle_diam = {cal_touching_circle_diam}")
            den = cal_touching_circle_diam / cal_diam - 1.0
            Q0 = m5 * (1.0 + 1.0 / den)

        elif self.res_type == "reflection_method2":
            if auto_flag:
                raise ValueError('Scaling factor must be defined for Method 2')
            cal_diam, S_V, S_T = self.Q_circle(opt_res, A)
            gv = abs(S_V)
            gv2 = gv * gv
            mb = abs(S_T)
            cosphi = (gv2 + cal_diam * cal_diam - mb * mb) / (
                2.0 * gv * cal_diam
            )  # Cosine rule
            cal_touching_circle_diam = (1.0 - gv2) / (1.0 - gv * cosphi)
            if self.verbose:
                print(f"Q-circle diam = {cal_diam}, touching_circle_diam = {cal_touching_circle_diam}")
            den = cal_touching_circle_diam / cal_diam - 1.0
            Q0 = m5 * (1.0 + 1.0 / den)

        elif self.res_type == "notch" or self.res_type == "absorption":  # By transmission
            if auto_flag:
                if self.verbose:
                    print(
                        'Notch/absorption Qo calculation: using fitted data to estimate scaling factor'
                    )
                # scale to S_V if A is undefined
                A = 1.0 / abs(complex(m1, m2))
            cal_diam, S_V, S_T = self.Q_circle(opt_res, A)
            if self.verbose:
                print(f"Q-circle diam = {cal_diam}")
            if cal_diam == 1.0:
                raise ZeroDivisionError("Divide by zero forestalled in calculation of Qo")
            den = 1.0 / cal_diam - 1.0  # Gao thesis (2008) 4.35 and 4.40
            Q0 = m5 * (
                1.0 + 1.0 / den
            )  # https://resolver.caltech.edu/CaltechETD:etd-06092008-235549
            # For this type of resonator, critical coupling occurs for cal_diam = 0.5.
        else:
            raise ValueError("Unknown resonance type {self.res_type}")

        return Q0

    def fitted_s(self,
                 opt_res: None | OptimizedResult = None,
                 f: None | np.ndarray = None
                 ) -> np.ndarray:
        r"""S-parameter response of an equivalent circuit model resonator.

        The approximate solution and estimated is obtained from [MAT58]_:

        .. math::

            S = m_1 + j m_2 + \frac{m_3 + j m_4}{1 + j Q_L t}

        where the m coefficients come from the fitted solution and

        .. math::

            t = \frac{f}{f_L} - \frac{f_L}{f}  \approx 2 \frac{f − f_L}{f_L}


        Parameters
        ----------
       opt_res : None or :class:`~skrf.qfactor.OptimizedResult`. Default is None.
           Solution produced by the :meth:`~skrf.qfactor.Qfactor.fit` method.
           If None, uses the solution previously calculated, if performed.
        f : None or np.ndarray. Default is None.
            frequency array [Hz]. If None, use the self frequencies.

        Returns
        -------
        s : np.ndarray
            S-parameter response.

        References
        ----------
        .. [MAT58] "Q-factor Measurement by using a Vector Network Analyser",
            A. P. Gregory, National Physical Laboratory Report MAT 58 (2021)
            https://eprintspublications.npl.co.uk/9304/,
            section 2.2, eqn. (21).

        """
        # if no solution passed, use internal solution if exist
        if opt_res is None:
            if self.opt_res is None:
                raise ValueError('No solution found or passed.')
            else:
                opt_res = self.opt_res

        if f is None:
            f = self.f

        # fractional offset frequency
        t = f/opt_res.f_L - opt_res.f_L/f

        y = 1/(1 + 1j*opt_res.Q_L*t)
        s = opt_res.m1 +1j*opt_res.m2 + (opt_res.m3 + 1j*opt_res.m4) * y
        return s

    def fitted_network(self,
                       opt_res: None | OptimizedResult = None,
                       frequency: None | Frequency = None,
                       ) -> Network:
        """Fitted Network.

        Return the Network corresponding to the fitted response.

        Parameters
        ----------
        opt_res : None or :class:`~skrf.qfactor.OptimizedResult`. Default is None.
            Solution produced by the :meth:`~skrf.qfactor.Qfactor.fit` method.
            If None, uses the solution previously calculated, if performed.
        frequency : None or :class:`~skrf.frequency.Frequency`. Default is None.
            Frequency for the fitted Network. If None, use the same
            Frequency than the one used to create the QFactor.

        Returns
        -------
        ntwk : :class:`~skrf.network.Network`
            Fitted Network for the passed Frequency.

        """
        # if no solution passed, use internal solution if exist
        if opt_res is None:
            if self.opt_res is None:
                raise ValueError('No solution found or passed.')
            else:
                opt_res = self.opt_res

        if frequency is None:
            frequency = self._ntwk.frequency

        s = self.fitted_s(opt_res, f=frequency.f)

        ntwk = Network(s=s, frequency=frequency)
        return ntwk

    @property
    def f_L_scaled(self) -> float:
        """
        Resonant Frequency in the frequency unit..

        Returns
        -------
        float
            Resonant frequency in the frequency unit.

        See Also
        --------
        f_L : Resonant Frequency in Hz.

        """
        if not self.fitted:
            warn('Q-factor not fitted, result may be inaccurate. Use the .fit() method before.', stacklevel=2)
        return self.f_L/self.f_multiplier

    @property
    def BW(self) -> float:
        r"""3-dB Bandwidth.

        Return the half-power fractional bandwidth (aka 3-dB bandwidth)
        defined as:

        .. math::

            BW = \frac{f_L}{Q_L}


        Returns
        -------
        float
            3-dB Bandwidth in Hz.

        See Also
        --------
        BW_scaled : 3-dB Bandwidth scaled to the frequency unit.

        """
        if not self.fitted:
            warn('Q-factor not fitted, result may be inaccurate. Use the .fit() method before.', stacklevel=2)
        return self.f_L/self.Q_L

    @property
    def BW_scaled(self) -> float:
        r"""3-dB Bandwidth scaled to the frequency unit.

        Returns
        -------
        float
            3-dB Bandwidth in the frequency unit.

        See Also
        --------
        BW : 3-dB Bandwidth in Hz.

        """
        if not self.fitted:
            warn('Q-factor not fitted, result may be inaccurate. Use the .fit() method before.', stacklevel=2)
        return self.BW/self.f_multiplier



import os
import unittest

import numpy as np

import skrf as rf
from skrf.frequency import Frequency
from skrf.media import MLine

try:
    from matplotlib import pyplot as plt
    rf.stylely()
except ImportError:
    pass



class MLineTestCase(unittest.TestCase):
    """
    Testcase for the MLine Media
    """
    def setUp(self):
        """
        Read in all the network data required for tests
        """
        self.data_dir_qucs = os.path.dirname(os.path.abspath(__file__)) + \
            '/qucs_prj/'
        self.data_dir_ads = os.path.dirname(os.path.abspath(__file__)) + \
            '/ads/'

        self.ref_qucs = [
            {'model': 'hammerstadjensen', 'disp': 'hammerstadjensen', 'color': 'r',
             'n': rf.Network(os.path.join(self.data_dir_qucs,
                               'mline,hammerstad,hammerstad.s2p'))},
            {'model': 'hammerstadjensen', 'disp': 'kirschningjansen', 'color': 'c',
             'n': rf.Network(os.path.join(self.data_dir_qucs,
                               'mline,hammerstad,kirschning.s2p'))},
            {'model': 'hammerstadjensen', 'disp': 'kobayashi', 'color': 'k',
             'n': rf.Network(os.path.join(self.data_dir_qucs,
                               'mline,hammerstad,kobayashi.s2p'))},
            {'model': 'hammerstadjensen', 'disp': 'yamashita', 'color': 'g',
             'n': rf.Network(os.path.join(self.data_dir_qucs,
                               'mline,hammerstad,yamashita.s2p'))},
            {'model': 'wheeler', 'disp': 'schneider', 'color': 'm',
             'n': rf.Network(os.path.join(self.data_dir_qucs,
                               'mline,wheeler,schneider.s2p'))},
            {'model': 'schneider', 'disp': 'schneider', 'color': 'b',
             'n': rf.Network(os.path.join(self.data_dir_qucs,
                               'mline,schneider,schneider.s2p'))}
            ]


        self.ref_ads = [
            {'diel': 'frequencyinvariant', 'disp': 'kirschningjansen', 'color': 'r',
             'n': rf.Network(os.path.join(self.data_dir_ads,
                           'mlin,freqencyinvariant,kirschning.s2p'))},
            {'diel': 'djordjevicsvensson', 'disp': 'kirschningjansen', 'color': 'c',
             'n': rf.Network(os.path.join(self.data_dir_ads,
                           'mlin,djordjevicsvensson,kirschning.s2p'))},
            {'diel': 'frequencyinvariant', 'disp': 'kobayashi', 'color': 'k',
             'n': rf.Network(os.path.join(self.data_dir_ads,
                           'mlin,freqencyinvariant,kobayashi.s2p'))},
            {'diel': 'djordjevicsvensson', 'disp': 'kobayashi', 'color': 'g',
             'n': rf.Network(os.path.join(self.data_dir_ads,
                           'mlin,djordjevicsvensson,kobayashi.s2p'))},
            {'diel': 'frequencyinvariant', 'disp': 'yamashita', 'color': 'm',
             'n': rf.Network(os.path.join(self.data_dir_ads,
                           'mlin,freqencyinvariant,yamashita.s2p'))},
            {'diel': 'djordjevicsvensson', 'disp': 'yamashita', 'color': 'b',
             'n': rf.Network(os.path.join(self.data_dir_ads,
                           'mlin,djordjevicsvensson,yamashita.s2p'))}
            ]

        # default parameter set for tests
        self.verbose = False # output comparison plots if True
        self.w    = 3.00e-3
        self.h    = 1.55e-3
        self.t    = 35e-6
        self.l    = 25e-3
        self.ep_r = 4.413
        self.tand = 0.0182
        self.rho  = 1.7e-8
        self.d    = 0.15e-6
        self.f_et = 1e9


    def test_z0_ep_reff(self):
        """
        Test against characterisitc impedance from another calculator using
        Hammerstadt-Jensen model
        http://web.mit.edu/~geda/arch/i386_rhel3/versions/20050830/html/mcalc-1.5/
        """
        freq = Frequency(1, 1, 1, 'GHz')
        mline1 = MLine(frequency = freq, z0_port = 50.,
                       w = self.w, h = self.h, t = self.t,
                       ep_r = self.ep_r, rho = self.rho,
                       tand = self.tand, rough = self.d,
                       diel = 'frequencyinvariant', disp = 'hammerstadjensen',
                       compatibility_mode = 'qucs')

        # without t (t = None)
        mline2 = MLine(frequency = freq, z0_port = 50.,
                       w = self.w, h = self.h,
                       ep_r = self.ep_r, rho = self.rho,
                       tand = self.tand, rough = self.d,
                       diel = 'frequencyinvariant', disp = 'hammerstadjensen',
                       compatibility_mode = 'qucs')

        # with t = 0
        mline3 = MLine(frequency = freq, z0_port = 50.,
                       w = self.w, h = self.h, t = 0,
                       ep_r = self.ep_r, rho = self.rho,
                       tand = self.tand, rough = self.d,
                       diel = 'frequencyinvariant', disp = 'hammerstadjensen',
                       compatibility_mode = 'qucs')

        self.assertTrue(np.abs((mline1.z0[0] - 49.142) / 49.142) < 0.01)
        self.assertTrue(np.abs((mline1.ep_reff_f[0] - 3.324) / 3.324) < 0.01)
        self.assertTrue(np.abs(mline2.w_eff - mline2.w) < 1e-16)
        self.assertTrue(np.abs(mline2.alpha_conductor) < 1e-16)
        self.assertTrue(np.abs(mline3.w_eff - mline3.w) < 1e-16)
        self.assertTrue(np.abs(mline3.alpha_conductor) < 1e-16)


    def test_line_qucs(self):
        """
        Test against the Qucs project results
        """
        if self.verbose:
            fig, axs = plt.subplots(2, 2, figsize = (8,6))
            fig.suptitle('qucs/skrf')
            fig2, axs2 = plt.subplots(2, 2, figsize = (8,6))
            fig2.suptitle('ads/skrf residuals')

        limit_db = 0.1
        limit_deg = 1.

        for ref in self.ref_qucs:
            mline = MLine(frequency = ref['n'].frequency, z0_port = 50.,
                            w = self.w, h = self.h, t = self.t,
                            ep_r = self.ep_r, rho = self.rho,
                            tand = self.tand, rough = self.d,
                            model = ref['model'], disp = ref['disp'],
                            diel = 'frequencyinvariant',
                            compatibility_mode = 'qucs')
            line = mline.line(d=self.l, unit='m')
            line.name = 'skrf,qucs'

            # residuals
            res = line / ref['n']
            res.name = 'residuals ' + ref['n'].name

            # test if within limit lines
            self.assertTrue(np.all(np.abs(res.s_db) < limit_db))
            self.assertTrue(np.all(np.abs(res.s_deg) < limit_deg))

            if self.verbose:
                line.plot_s_db(0, 0, ax = axs[0, 0], color = ref['color'],
                               linestyle = 'none', marker = 'x')
                ref['n'].plot_s_db(0, 0, ax = axs[0, 0], color = ref['color'])
                res.plot_s_db(0, 0, ax = axs2[0, 0], linestyle = 'dashed',
                              color = ref['color'])

                line.plot_s_deg(0, 0, ax = axs[0, 1], color = ref['color'],
                               linestyle = 'none', marker = 'x')
                ref['n'].plot_s_deg(0, 0, ax = axs[0, 1], color = ref['color'])
                res.plot_s_deg(0, 0, ax = axs2[0, 1], linestyle = 'dashed',
                              color = ref['color'])

                line.plot_s_db(1, 0, ax = axs[1, 0], color = ref['color'],
                               linestyle = 'none', marker = 'x')
                ref['n'].plot_s_db(1, 0, ax = axs[1, 0], color = ref['color'])
                res.plot_s_db(1, 0, ax = axs2[1, 0], linestyle = 'dashed',
                              color = ref['color'])

                line.plot_s_deg(1, 0, ax = axs[1, 1], color = ref['color'],
                               linestyle = 'none', marker = 'x')
                ref['n'].plot_s_deg(1, 0, ax = axs[1, 1], color = ref['color'])
                res.plot_s_deg(1, 0, ax = axs2[1, 1], linestyle = 'dashed',
                              color = ref['color'])


        if self.verbose:
            axs[1, 0].legend(prop={'size': 6})
            axs[0, 0].get_legend().remove()
            axs[0, 1].get_legend().remove()
            axs[1, 1].get_legend().remove()
            fig.tight_layout()

            axs2[1, 0].legend(prop={'size': 6})
            axs2[0, 0].get_legend().remove()
            axs2[0, 1].get_legend().remove()
            axs2[1, 1].get_legend().remove()
            fig2.tight_layout()

    def test_line_ads(self):
        """
        Test against the ADS results
        """
        if self.verbose:
            fig, axs = plt.subplots(2, 2, figsize = (8,6))
            fig.suptitle('ads/skrf')
            fig2, axs2 = plt.subplots(2, 2, figsize = (8,6))
            fig2.suptitle('ads/skrf residuals')

        # todo: restore to smal values
        limit_db = 0.1
        limit_deg = 1.

        for ref in self.ref_ads:
            mline = MLine(frequency = ref['n'].frequency, z0_port = 50.,
                            w = self.w, h = self.h, t = self.t,
                            ep_r = self.ep_r, rho = self.rho,
                            tand = self.tand, rough = self.d,
                            model = 'hammerstadjensen', disp = ref['disp'],
                            diel = ref['diel'])
            line = mline.line(d=self.l, unit='m')
            line.name = 'skrf,ads'

            # residuals
            res = line / ref['n']
            res.name = 'residuals ' + ref['n'].name

            # test if within limit lines
            # fixme: still a small deviation of S11 at low frequency
            #        limit line multiplied by 10 for S11 as for now
            self.assertTrue(
                np.all(np.abs(res.s_db[:, 0, 0]) < 10. *limit_db))
            self.assertTrue(
                np.all(np.abs(res.s_deg[:, 0, 0]) < 10. * limit_deg))
            self.assertTrue(np.all(np.abs(res.s_db[:, 1, 0]) < limit_db))
            self.assertTrue(np.all(np.abs(res.s_deg[:, 1, 0]) < limit_deg))

            if self.verbose:
                line.plot_s_db(0, 0, ax = axs[0, 0], color = ref['color'],
                               linestyle = 'none', marker = 'x')
                ref['n'].plot_s_db(0, 0, ax = axs[0, 0], color = ref['color'])
                res.plot_s_db(0, 0, ax = axs2[0, 0], linestyle = 'dashed',
                              color = ref['color'])

                line.plot_s_deg(0, 0, ax = axs[0, 1], color = ref['color'],
                               linestyle = 'none', marker = 'x')
                ref['n'].plot_s_deg(0, 0, ax = axs[0, 1], color = ref['color'])
                res.plot_s_deg(0, 0, ax = axs2[0, 1], linestyle = 'dashed',
                              color = ref['color'])

                line.plot_s_db(1, 0, ax = axs[1, 0], color = ref['color'],
                               linestyle = 'none', marker = 'x')
                ref['n'].plot_s_db(1, 0, ax = axs[1, 0], color = ref['color'])
                res.plot_s_db(1, 0, ax = axs2[1, 0], linestyle = 'dashed',
                              color = ref['color'])

                line.plot_s_deg(1, 0, ax = axs[1, 1], color = ref['color'],
                               linestyle = 'none', marker = 'x')
                ref['n'].plot_s_deg(1, 0, ax = axs[1, 1], color = ref['color'])
                res.plot_s_deg(1, 0, ax = axs2[1, 1], linestyle = 'dashed',
                              color = ref['color'])


        if self.verbose:
            axs[1, 0].legend(prop={'size': 6})
            axs[0, 0].get_legend().remove()
            axs[0, 1].get_legend().remove()
            axs[1, 1].get_legend().remove()
            fig.tight_layout()

            axs2[1, 0].legend(prop={'size': 6})
            axs2[0, 0].get_legend().remove()
            axs2[0, 1].get_legend().remove()
            axs2[1, 1].get_legend().remove()
            fig2.tight_layout()

    def test_alpha_warning(self):
        """
        Test if warns when t < 3 * skin_depth
        """
        freq = Frequency(1, 1, 1, 'MHz')
        with self.assertWarns(RuntimeWarning) as context:
            mline = MLine(frequency = freq, z0_port = 50.,
                       w = self.w, h = self.h, t = self.t,
                       ep_r = self.ep_r, rho = self.rho,
                       tand = self.tand, rough = self.d,
                       diel = 'frequencyinvariant', disp = 'hammerstadjensen')



"""
cpw (:mod:`skrf.media.cpw`)
========================================

.. autosummary::
   :toctree: generated/

   CPW

"""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from numpy import any, exp, imag, log, real, sinh, sqrt, tanh, zeros
from scipy.constants import c, epsilon_0, mu_0, pi
from scipy.special import ellipk

from ..constants import NumberLike
from ..tlineFunctions import skin_depth, surface_resistivity
from .media import Media

if TYPE_CHECKING:
    from ..frequency import Frequency


class CPW(Media):
    r"""
    Coplanar waveguide.


    A coplanar waveguide transmission line is defined in terms of width,
    spacing, and thickness on a given relative permittivity substrate of a
    certain height. The line has a conductor resistivity and a tangential loss
    factor. The backside of the strip can be made of air or metal (grounded
    coplanar waveguide).

    This class is highly inspired by the technical documentation [QUCSa]_
    and sources provided by the qucs project [QUCSb]_ .

    In addition, Djordjevic [DBLS01]_ /Svensson [SvDe01]_  wideband debye dielectric
    model is considered to provide a more realistic modelling of broadband
    microstrip with causal time domain response.

    A compatibility mode is provided to mimic the behaviour of QUCS or of
    Keysight ADS. There is known differences in the output of these
    simulators.

    The quasi-static models of characteristic impedance and effective
    permittivity give the value at zero frequency. The dispersion models
    compute frequency-dependant values of these variables.

    * Quasi-static characteristic impedance and effective permittivity model
      use [GhNa84]_ and [GhNa83]_. The models are corrected to account for
      strip thickness using a first-order approach described in [GGBB96]_.
      A comparison shows that ADS simulator uses another thickness correction
      method that is according to ADS doc based on [Cohn60]_. This second method
      is not implemented in skrf.

    * Frequency dispersion of impedance and effective permittivity model use
      [FGVM91]_ and [GMDK97]_.

    * Loss model is computed using Wheeler's incremental inductance rule
      [Whee42]_ applied to coplanar waveguide by [OwWu58]_ and [Ghio93]_.

    Parameters
    ----------
    frequency : :class:`~skrf.frequency.Frequency` object, optional
        frequency band of the media. The default is None.
    z0_port : number, array-like, or None
        `z0_port` is the port impedance for networks generated by the media.
        If `z0_port` is not None, the networks generated by the media are
        renormalized (or in other words embedded) from the characteristic
        impedance z0 of the media to `z0_port`.
        Else if `z0_port` is None, the networks port impedances will be the raw
        characteristic impedance z0 of the media.
        (Default is None)
    z0_override : number, array-like, or None
        `z0_override` override the characteristic impedance for the media.
        If `z0_override` is not None, the networks generated by the media have
        their characteristic impedance `z0` overrided by `z0_override`.
        (Default is None)
    z0 : number, array-like, or None
        deprecated parameter, alias to `z0_override` if `z0_override` is None.
        Emmit a deprecation warning.
    w : number, or array-like
        width of the center conductor, in m. Default is 3e-3 m.
    s : number, or array-like
        spacing (width of the gap), in m. Default is 0.3e-3 m.
    h : number, or array-like
        height of the substrate between backside and conductor, in m.
        Default is 1.55 m (equivalent to infinite height for default w and s).
    t : number, or array-like, optional
        conductor thickness, in m. Default is None (no width correction
        to account for strip thickness).
    has_metal_backside : bool, default False
        If the backside is air (False) or metal (True)
    ep_r : number, or array-like, optional
        relative permittivity of the substrate at frequency f_epr_tand,
        no unit. Default is 4.5.
    diel : str
        dielectric frequency dispersion model in:

        * 'djordjevicsvensson' (default)
        * 'frequencyinvariant'

    rho : number, or array-like, or None
        resistivity of conductor, ohm / m. Default is 1.68e-8 ohm /m (copper).
    tand : number, or array-like
        dielectric loss factor at frequency f_epr_tand. Default is 0.
    f_low : number, or array-like
        lower frequency for wideband Debye Djordjevic/Svensson dielectric
        model, in Hz. Default is 1 kHz.
    f_high : number, or array-like
        higher frequency for wideband Debye Djordjevic/Svensson dielectric
        model, in Hz. Default is 1 THz.
    f_epr_tand : number, or array-like
        measurement frequency for ep_r and tand of dielectric, in Hz.
        Default is 1 GHz.
    compatibility_mode: str or None (default)
        If set to 'qucs', following behaviour happens :

        * Characteristic impedance will be real (no imaginary part due to tand)

    \*args, \*\*kwargs : arguments, keyword arguments
            passed to :class:`~skrf.media.media.Media`'s constructor
            (:func:`~skrf.media.media.Media.__init__`

    Note
    ----
    When the thickness of the strip is smaller than 3 skin depth, the losses
    model gives over-optimistic results and the media will issue a warning.
    At DC, the losses of the line could be smaller than its conductor
    resistance, which is not physical.

    References
    ----------
    .. [QUCSa] http://qucs.sourceforge.net/docs/technical.pdf
    .. [QUCSb] http://www.qucs.sourceforge.net/
    .. [DBLS01] Djordjevic, R.M. Biljic, V.D. Likar-Smiljanic, T.K. Sarkar,
        Wideband frequency-domain characterization of FR-4 and time-domain
        causality,
        IEEE Trans. on EMC, vol. 43, N4, 2001, p. 662-667.
    .. [SvDe01] C. Svensson, G.E. Dermer,
        Time domain modeling of lossy interconnects,
        IEEE Trans. on Advanced Packaging, May 2001, N2, Vol. 24, pp.191-196.
    .. [GhNa84] G. Ghione and C. Naldi. "Analytical Formulas for Coplanar Lines
       in Hybrid and Monolithic MICs", Electronics Letters,
       Vol. 20, No. 4, February 16, 1984, pp. 179-181.
    .. [GhNa83] G. Ghione and C. Naldi. "Parameters of Coplanar Waveguides with
        Lower Common Planes", Electronics Letters,
        Vol. 19, No. 18, September 1, 1983, pp. 734-735.
    .. [Cohn60] S. B. Cohn, "Thickness Corrections for Capacitive obstacles and
       Strip Conductors", IRE Trans. on Microwave Theory and Techniques,
       Vol. MTT-8, November 1960, pp. 638-644.
    .. [GGBB96] K. C. Gupta, R. Garg, I. J. Bahl, and P. Bhartia, Microstrip
       Lines and Slotlines, 2nd ed.Artech House, Inc., 1996.
    .. [FGVM91] M. Y. Frankel, S. Gupta, J. A. Valdmanis, and G. A. Mourou,
       "Terahertz Attenuation and Dispersion Characteristics of Coplanar
       Transmission Lines" IEEE Trans. on Microwave Theory and Techniques,
       vol. 39, no. 6, pp. 910-916, June 1991.
    .. [GMDK97] S. Gevorgian, T. Martinsson, A. Deleniv, E. Kollberg, and
       I. Vendik, "Simple and accurate dispersion expression for the
       effective dielectric constant of coplanar waveguides" in
       Proceedings of Microwaves, Antennas and Propagation,
       vol. 144, no. 2.IEE, Apr. 1997, pp. 145-148.
    .. [Whee42] H. A. Wheeler, "Formulas for the Skin Effect,"
        Proceedings of the IRE, vol. 30, no. 9, pp. 412-424, Sept. 1942.
    .. [OwWu58] G. H. Owyang and T. T. Wu, "The Approximate Parameters of Slot
        Lines and Their Complement" IRE Transactions on Antennas and
        Propagation, pp. 49-55, Jan. 1958.
    .. [Ghio93] G. Ghione, "A CAD-Oriented Analytical Model for the Losses of
        General Asymmetric Coplanar Lines in Hybrid and Monolithic MICs"
        IEEE Trans. on Microwave Theory and Techniques,
        vol. 41, no. 9, pp. 1499-1510, Sept. 1993.

    """
    def __init__(self, frequency: Frequency | None = None,
                 z0_port: NumberLike | None = None,
                 z0_override: NumberLike | None = None,
                 z0: NumberLike | None = None,
                 w: NumberLike = 3e-3, s: NumberLike = 0.3e-3,
                 h: NumberLike = 1.55,
                 ep_r: NumberLike = 4.5, t: NumberLike | None = None,
                 diel: str = 'djordjevicsvensson',
                 rho: NumberLike | None = 1.68e-8, tand: NumberLike = 0,
                 f_low: NumberLike = 1e3, f_high: NumberLike = 1e12,
                 f_epr_tand: NumberLike = 1e9,
                 has_metal_backside: bool = False,
                 compatibility_mode: str | None = None,
                 *args, **kwargs):
        Media.__init__(self, frequency = frequency,
                       z0_port = z0_port, z0_override = z0_override, z0 = z0)

        self.w, self.s, self.h, self.t, self.ep_r, self.tand, self.rho =\
                w, s, h, t, ep_r, tand, rho
        self.diel = diel
        self.f_low, self.f_high, self.f_epr_tand = f_low, f_high, f_epr_tand
        self.has_metal_backside = has_metal_backside
        self.compatibility_mode = compatibility_mode

        # variation of effective permittivity with frequency
        # Not implemented in QUCS but implemented in ADS.
        # 'frequencyinvariant' will give a constant complex value whith a real
        # part compatible with qucs and an imaginary part due to tand
        self.ep_r_f, self.tand_f = self.analyse_dielectric(
            self.ep_r, self.tand,
            self.f_low, self.f_high, self.f_epr_tand, self.frequency.f,
            self.diel)

        # quasi-static effective permittivity of substrate + line and
        # the impedance of the coplanar waveguide
        # qucs use real-valued ep_r giving real-valued impedance
        if compatibility_mode == 'qucs':
            self.zl_eff, self.ep_reff, k1, kk1, kpk1 = \
                self.analyse_quasi_static(
                real(self.ep_r_f), w, s, h, t, has_metal_backside)
        else:
            self.zl_eff, self.ep_reff, k1, kk1, kpk1 = \
                self.analyse_quasi_static(
                self.ep_r_f, w, s, h, t, has_metal_backside)

        # analyse dispersion of impedance and relatice permittivity
        if compatibility_mode == 'qucs':
            self._z_characteristic, self.ep_reff_f = self.analyse_dispersion(
                self.zl_eff, self.ep_reff, real(self.ep_r_f), w, s, h,
                self.frequency.f)
        # ads does not use frequency dispersion for cpw/cpwg (so sad)
        elif compatibility_mode == 'ads':
            self._z_characteristic = self.zl_eff
            self.ep_reff_f = self.ep_reff
        # use frequency dispersion and complex permittivity by default
        else:
            self._z_characteristic, self.ep_reff_f = self.analyse_dispersion(
                self.zl_eff, self.ep_reff, self.ep_r_f, w, s, h,
                self.frequency.f)

        # analyse losses of line
        self.alpha_conductor, self.alpha_dielectric = self.analyse_loss(
            real(self.ep_r_f), real(self.ep_reff_f), self.tand_f, rho, 1.,
            self.frequency.f,
            w, t, s, k1, kk1, kpk1)

    def __str__(self) -> str:
        f=self.frequency
        output = (f'Coplanar Waveguide Media.  {f.f_scaled[0]}-{f.f_scaled[-1]} {f.unit}. {f.npoints} points'
                  f'\n W= {self.w:.2e}m, S= {self.s:.2e}m')
        return output

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def z0_characteristic(self) -> NumberLike:
        """
        Characteristic Impedance, :math:`z_0`.

        Returns
        -------
        z0_characteristic : np.ndarray
            Characteristic Impedance in units of ohms
        """
        return self._z_characteristic

    @property
    def gamma(self) -> NumberLike:
        """
        Propagation constant.

        Returns
        -------
        gamma : :class:`numpy.ndarray`
        """
        ep_reff, f = real(self.ep_reff_f), self.frequency.f

        alpha = self.alpha_dielectric.copy()
        if self.rho is not None:
            alpha += self.alpha_conductor

        beta = 2. * pi * f * sqrt(ep_reff) / c

        return alpha + 1j * beta

    def analyse_dielectric(self, ep_r: NumberLike, tand: NumberLike,
                          f_low: NumberLike, f_high: NumberLike,
                          f_epr_tand: NumberLike, f: NumberLike,
                          diel: str):
        """
        This function calculate the frequency-dependent relative permittivity
        of dielectric and tangential loss factor.

        References
        ----------
        .. [#] C. Svensson, G.E. Dermer,
            Time domain modeling of lossy interconnects,
            IEEE Trans. on Advanced Packaging, May 2001, N2, Vol. 24, pp.191-196.
        .. [#] Djordjevic, R.M. Biljic, V.D. Likar-Smiljanic, T.K. Sarkar,
            Wideband frequency-domain characterization of FR-4 and time-domain
            causality,
            IEEE Trans. on EMC, vol. 43, N4, 2001, p. 662-667.

        Returns
        -------
        ep_r_f : :class:`numpy.ndarray`
        tand_f : :class:`numpy.ndarray`
        """
        if diel == 'djordjevicsvensson':
            # compute the slope for a log frequency scale, tanD dependent.
            k = log((f_high + 1j * f_epr_tand) / (f_low + 1j * f_epr_tand))
            fd = log((f_high + 1j * f) / (f_low + 1j * f))
            ep_d = -tand * ep_r  / imag(k)
            # value for frequency above f_high
            ep_inf = ep_r * (1. + tand * real(k) / imag(k))
            # compute complex permitivity
            ep_r_f = ep_inf + ep_d * fd
            # get tand
            tand_f = -imag(ep_r_f) / real(ep_r_f)
        elif diel == 'frequencyinvariant':
            ep_r_f =  ep_r - 1j * ep_r * tand
            tand_f = tand
        else:
            raise ValueError('Unknown dielectric dispersion model')

        return ep_r_f, tand_f

    def analyse_quasi_static(self, ep_r: NumberLike,
                           w: NumberLike, s: NumberLike,
                           h: NumberLike, t: NumberLike,
                           has_metal_backside: bool):
        """
        This function calculates the quasi-static impedance of a coplanar
        waveguide line, the value of the effective permittivity as per filling
        factor, and the effective width due to the finite conductor thickness
        for the given coplanar waveguide line and substrate properties.
        Model from [#]_ with air backside and [#]_ with a metal backside.
        The models are corrected to account for
        strip thickness using a first-order approach described in [GGBB96]_.
        ADS simulator report to use a custom correction based on [Cohn60]_.
        This second method is not implemented in skrf.

        References
        ----------
        .. [GhNa84] G. Ghione and C. Naldi. "Analytical Formulas for Coplanar Lines
           in Hybrid and Monolithic MICs", Electronics Letters,
           Vol. 20, No. 4, February 16, 1984, pp. 179-181.
        .. [GhNa83] G. Ghione and C. Naldi. "Parameters of Coplanar Waveguides with
            Lower Common Planes", Electronics Letters,
            Vol. 19, No. 18, September 1, 1983, pp. 734-735.
        .. [Cohn60] S. B. Cohn, "Thickness Corrections for Capacitive obstacles and
           Strip Conductors", IRE Trans. on Microwave Theory and Techniques,
           Vol. MTT-8, November 1960, pp. 638-644.
        .. [GGBB96] K. C. Gupta, R. Garg, I. J. Bahl, and P. Bhartia, Microstrip
           Lines and Slotlines, 2nd ed.Artech House, Inc., 1996.

        Returns
        -------
        zl_eff : :class:`numpy.ndarray`
        ep_reff : :class:`numpy.ndarray`
        """
        Z0 = sqrt(mu_0 / epsilon_0)
        a = w
        b = w + 2. * s

        # equation (3a) from [GhNa84] or (6) from [GhNa83]
        k1 = a / b
        kk1 = ellipk(k1)
        kpk1 = ellipk(sqrt(1. - k1 * k1))
        q1 = ellipa(k1)

        # backside is metal
        if has_metal_backside:
            # equation (4) from [GhNa83]
            # in qucs the 2 coefficient turn to 4 and fit better with ads
            k3 = tanh(pi * a / 4. / h) / tanh(pi * b / 4. / h)
            q3 = ellipa(k3)
            qz = 1. / (q1 + q3)
            # equation (7) from [GhNa83]
            # equivalent to e = (q1 + ep_r * q3) / (q1 + q3) and paper
            e = 1. + q3 * qz * (ep_r - 1.)
            # equation (8) from [GhNa83] with later division by sqrt(e)
            zr = Z0 / 2. * qz

        # backside is air
        else:
            # equation (3b) from [GhNa84]
            k2 = sinh((pi / 4.) * a / h) / sinh((pi / 4.) * b / h)
            q2 = ellipa(k2)
            # equation (2) from [GhNa84]
            e = 1. + (ep_r - 1.) / 2. * q2 / q1
            # equation (1) from [GhNa84] with later division by sqrt(e)
            zr = Z0 / 4. / q1

        # a posteriori effect of strip thickness
        if t is not None and t > 0.:
            # equation (7.98) from [GGBB96]
            d = 1.25 * t / pi * (1. + log(4. * pi * w / t))
            # equation between (7.99) and (7.100) from [GGBB96]
            #approx. equal to ke = (w + d) / (w + d + 2 * (s - d))
            ke = k1 + (1. - k1 * k1) * d / 2. / s
            qe = ellipa(ke)

            # backside is metal
            if has_metal_backside:
                # equation (8) from [GhNa83] with k1 -> ke
                # but keep q3 unchanged ? (not in papers)
                qz = 1. / (qe + q3)
                zr = Z0 / 2. * qz
            # backside is air
            else:
                # equation (7.99) from [GGBB96] with later division by sqrt(e)
                zr = Z0 / 4. / qe

            # modifies ep_re
            # equation (7.100) of [GGBB96]
            e = e - (0.7 * (e - 1.) * t / s) / (q1 + (0.7 * t / s))

        ep_reff = e
        # final division of (1) from [GhNa84] and (8) from [GhNa83]
        zl_eff = zr / sqrt(ep_reff)

        return zl_eff, ep_reff, k1, kk1, kpk1

    def analyse_dispersion(self, zl_eff: NumberLike, ep_reff: NumberLike,
                          ep_r: NumberLike, w: NumberLike, s: NumberLike,
                          h: NumberLike, f: NumberLike):
         """
         This function computes the frequency-dependent characteristic
         impedance and effective permittivity accounting for coplanar waveguide
         frequency dispersion.

         References
         ----------
         .. [#] M. Y. Frankel, S. Gupta, J. A. Valdmanis, and G. A. Mourou,
            "Terahertz Attenuation and Dispersion Characteristics of Coplanar
            Transmission Lines" IEEE Trans. on Microwave Theory and Techniques,
            vol. 39, no. 6, pp. 910-916, June 1991.
         .. [#] S. Gevorgian, T. Martinsson, A. Deleniv, E. Kollberg, and
            I. Vendik, "Simple and accurate dispersion expression for the
            effective dielectric constant of coplanar waveguides" in
            Proceedings of Microwaves, Antennas and Propagation,
            vol. 144, no. 2.IEE, Apr. 1997, pp. 145-148.

         Returns
         -------
         z : :class:`numpy.ndarray`
         e : :class:`numpy.ndarray`
         """
         # cut-off frequency of the TE0 mode
         fte = ((c / 4.) / (h * sqrt(ep_r - 1.)))

         # dispersion factor G
         p = log(w / h)
         u = 0.54 - (0.64 - 0.015 * p) * p
         v = 0.43 - (0.86 - 0.54 * p) * p
         G = exp(u * log(w / s) + v)

         # add the dispersive effects to ep_reff
         sqrt_ep_reff = sqrt(ep_reff)
         sqrt_e = sqrt_ep_reff + (sqrt(ep_r) - sqrt_ep_reff) / \
             (1. + G * (f / fte)**(-1.8))

         e = sqrt_e**2

         z = zl_eff * sqrt_ep_reff / sqrt_e

         return z, e

    def analyse_loss(self, ep_r: NumberLike, ep_reff: NumberLike,
                    tand: NumberLike, rho: NumberLike, mu_r: NumberLike,
                    f: NumberLike, w: NumberLike, t: NumberLike, s: NumberLike,
                    k1: NumberLike, kk1: NumberLike, kpk1: NumberLike):
        """
        The function calculates the conductor and dielectric losses of a
        complanar waveguide line using wheeler's incremental inductance rule.

        References
        ----------
        .. [#] H. A. Wheeler, "Formulas for the Skin Effect,"
            Proceedings of the IRE, vol. 30, no. 9, pp. 412-424, Sept. 1942.
        .. [#] G. H. Owyang and T. T. Wu, "The Approximate Parameters of Slot
            Lines and Their Complement" IRE Transactions on Antennas and
            Propagation, pp. 49-55, Jan. 1958.
        .. [#] G. Ghione, "A CAD-Oriented Analytical Model for the Losses of
            General Asymmetric Coplanar Lines in Hybrid and Monolithic MICs"
            IEEE Trans. on Microwave Theory and Techniques,
            vol. 41, no. 9, pp. 1499-1510, Sept. 1993.

        Returns
        -------
        a_conductor : :class:`numpy.ndarray`
        a_dielectric : :class:`numpy.ndarray`
        """
        Z0 = sqrt(mu_0 / epsilon_0)
        if t is not None and t > 0.:
            if rho is None:
                raise(AttributeError("must provide values conductivity and conductor thickness to calculate this. "
                                     "see initializer help"))
            r_s = surface_resistivity(f=f, rho=rho, \
                    mu_r=1)
            ds = skin_depth(f = f, rho = rho, mu_r = 1.)
            if any(t < 3 * ds):
                warnings.warn(
                    'Conductor loss calculation invalid for line'
                    f'height t ({t})  < 3 * skin depth ({ds[0]})',
                    RuntimeWarning, stacklevel=2
                    )
            n = (1. - k1) * 8. * pi / (t * (1. + k1))
            a = w / 2.
            b = a + s
            ac = (pi + log(n * a)) / a + (pi + log(n * b)) / b
            a_conductor = r_s * sqrt(ep_reff) * ac / (4. * Z0 * kk1 * kpk1 * \
                               (1. - k1 * k1))
        else:
            a_conductor = zeros(f.shape)

        l0 = c / f
        a_dielectric =  pi * ep_r / (ep_r - 1) * (ep_reff - 1) / \
            sqrt(ep_reff) * tand / l0

        return a_conductor, a_dielectric

def ellipa(k: NumberLike):
    """
    Approximation of K(k)/K'(k).
    First appeared in [#]_
    More accurate expressions can be found in the above article and in [#]_.

    The maximum relative error of the approximation implemented here is
    about 2 ppm, so good enough for any practical purpose.

    References
    ==========
    .. [#] Hilberg, W., "From Approximations to Exact Relations for
       Characteristic Impedances," IEEE Trans. MTT, May 1969.
    .. [#] Abbott, J. T., "Modeling the Capacitive Behavior of Coplanar
       Striplines and Coplanar Waveguides using Simple Functions",
       Rochester Institute of Technology, Rochester, New York, June 2011.
    """
    if k < sqrt(1. / 2.):
        kp = sqrt(1 - k * k)
        r = pi / log(2. * (1. + sqrt(kp)) / (1. - sqrt(kp)))
    else:
        r = log(2. * (1. + sqrt(k)) / (1. - sqrt(k))) / pi

    return r



import sys

try:
    import matplotlib as mpl
except ImportError:
    pass

import numpy as np
import pytest

import skrf as rf

if "matplotlib" not in sys.modules:
    pytest.skip(allow_module_level=True)


@pytest.fixture(params=rf.Network.PRIMARY_PROPERTIES)
def primary_properties(request):
    return request.param

@pytest.fixture(params=["polar", "complex"])
def primary_methods(request):
    return request.param

@pytest.fixture(params=rf.Network._generated_functions().keys())
def generated_functions(request):
    return request.param

def test_primary_plotting(ntwk1_dc: rf.Network, primary_properties: str, primary_methods: str):
    method = f"plot_{primary_properties}_{primary_methods}"
    fig = getattr(ntwk1_dc, method)()

def test_generated_function_plots(ntwk1_dc: rf.Network, generated_functions: str):
    method = f"plot_{generated_functions}"
    fig = getattr(ntwk1_dc, method)()

def test_plot_passivity(ntwk1_dc: rf.Network):
    ntwk1_dc.plot_passivity()

def test_plot_reciprocity(ntwk1_dc: rf.Network, ):
    ntwk1_dc.plot_reciprocity()

def test_plot_reciprocity2(ntwk1_dc: rf.Network, ):
    ntwk1_dc.plot_reciprocity2()

def test_plot_s_db_time(ntwk1_dc: rf.Network, ):
    ntwk1_dc.plot_s_db_time()
    ntwk1_dc.plot_s_db_time(m=0, n=0)

def test_plot_s_smith(ntwk1_dc: rf.Network, ):
    ntwk1_dc.plot_s_smith()

def test_z_time_impulse_step_z0(ntwk1_dc: rf.Network, ):
    se_diff = rf.concat_ports([ntwk1_dc, ntwk1_dc], port_order = 'second')
    mm_diff = se_diff.copy()
    mm_diff.se2gmm(p = 2)
    fig, ax = mpl.pyplot.subplots(1, 1)
    mm_diff.plot_z_time_step(0, 0, ax = ax)
    mm_diff.plot_z_time_impulse(0, 1, ax = ax)
    mm_diff.plot_z_time_step(2, 2, ax = ax)
    mm_diff.plot_z_time_impulse(2, 3, ax = ax)
    # test that start impedance is almost port z0 to 0.1%
    z0_start = [
        ax.lines[0].get_ydata()[0],
        ax.lines[1].get_ydata()[0],
        ax.lines[2].get_ydata()[0],
        ax.lines[3].get_ydata()[0]]
    np.testing.assert_allclose(z0_start, mm_diff.z0[0],
        rtol = 1e-3,
        err_msg = "plot_z_time_xxx start impedance does not match port impedance")

def test_x_y_labels(ntwk1_dc: rf.Network):
    fig, ax = mpl.pyplot.subplots(1, 1)
    ntwk1_dc.plot_s_db(ax = ax)
    # check that there are 4 traces
    np.testing.assert_equal(len(ax.lines), ntwk1_dc.nports**2,
                                    err_msg = "missing trace")
    # check x label
    np.testing.assert_string_equal(ax.get_xlabel(), f"Frequency ({ntwk1_dc.frequency.unit})")
    # check y label
    np.testing.assert_string_equal(ax.get_ylabel(), rf.Network.Y_LABEL_DICT['db'])



@pytest.mark.parametrize("usetex", [True, False])
def test_plot_it_all(ntwk1_dc: rf.Network, usetex):
    with mpl.rc_context({'text.usetex': usetex}):
        ntwk1_dc.plot_it_all()

@pytest.mark.parametrize("usetex", [True, False])
def test_plot_polar(ntwk1_dc: rf.Network, usetex):
    with mpl.rc_context({'text.usetex': usetex}):
        ntwk1_dc.plot_s_polar()

def test_plot_uncertainty_bounds_s(ntwk_set_zip: rf.NetworkSet):
    ntwk_set_zip.plot_uncertainty_bounds_s()

def test_plot_uncertainty_bounds_s_db(ntwk_set_zip: rf.NetworkSet):
    ntwk_set_zip.plot_uncertainty_bounds_s_db()

def test_plot_uncertainty_bounds_s_time_db(ntwk_set_zip: rf.NetworkSet):
    ntwk_set_zip.plot_uncertainty_bounds_s_time_db()

def test_plot_uncertainty_decomposition(ntwk_set_zip: rf.NetworkSet):
    ntwk_set_zip.plot_uncertainty_decomposition()

def test_plot_minmax_bounds_s_db(ntwk_set_zip: rf.NetworkSet):
    ntwk_set_zip.plot_minmax_bounds_s_db()

def test_plot_minmax_bounds_s_db10(ntwk_set_zip: rf.NetworkSet):
    ntwk_set_zip.plot_minmax_bounds_s_db10()

def test_plot_minmax_bounds_s_time_db(ntwk_set_zip: rf.NetworkSet):
    ntwk_set_zip.plot_minmax_bounds_s_time_db()

def test_plot_signature(ntwk_set_zip: rf.NetworkSet):
    ntwk_set_zip.signature()

def test_generated_violin_plots(ntwk_set_zip: rf.NetworkSet, generated_functions: str):
    if "time" not in generated_functions:
        fig = ntwk_set_zip.plot_violin(generated_functions)
    else:
        with pytest.raises(NotImplementedError):
            fig = ntwk_set_zip.plot_violin(generated_functions)
